#   Copyright 2014 CERN.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstack.identity.v3 import domain as _domain
from openstack.identity.v3 import identity_provider as _identity_provider
from openstack.test import fakes as sdk_fakes
from openstackclient.identity.v3 import identity_provider
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit import utils as test_utils


class TestIdentityProviderCreate(identity_fakes.TestFederatedIdentity):
    columns = (
        'authorization_ttl',
        'description',
        'domain_id',
        'enabled',
        'id',
        'remote_ids',
    )

    def setUp(self):
        super().setUp()

        self.domain = sdk_fakes.generate_fake_resource(_domain.Domain)
        self.idp = sdk_fakes.generate_fake_resource(
            _identity_provider.IdentityProvider,
            domain_id=self.domain.id,
            remote_ids=['entity1', 'entity2'],
        )

        self.identity_sdk_client.create_identity_provider.return_value = (
            self.idp
        )
        self.identity_sdk_client.find_domain.return_value = self.domain

        self.datalist = (
            self.idp.authorization_ttl,
            self.idp.description,
            self.domain.id,
            self.idp.is_enabled,
            self.idp.id,
            format_columns.ListColumn(self.idp.remote_ids),
        )
        self.cmd = identity_provider.CreateIdentityProvider(self.app, None)

    def test_create_identity_provider_no_options(self):
        arglist = [
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider_id', self.idp.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
        }

        self.identity_sdk_client.create_identity_provider.assert_called_with(
            id=self.idp.id, **kwargs
        )

        self.assertEqual(self.columns, columns)

        self.assertCountEqual(self.datalist, data)

    def test_create_identity_provider_description(self):
        arglist = [
            '--description',
            self.idp.description,
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider_id', self.idp.id),
            ('description', self.idp.description),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': self.idp.description,
            'is_enabled': True,
        }

        self.identity_sdk_client.create_identity_provider.assert_called_with(
            id=self.idp.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)

    def test_create_identity_provider_remote_id(self):
        arglist = [
            self.idp.id,
            '--remote-id',
            self.idp.remote_ids[0],
        ]
        verifylist = [
            ('identity_provider_id', self.idp.id),
            ('remote_ids', self.idp.remote_ids[:1]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'remote_ids': self.idp.remote_ids[:1],
            'is_enabled': True,
        }

        self.identity_sdk_client.create_identity_provider.assert_called_with(
            id=self.idp.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)

    def test_create_identity_provider_remote_ids_multiple(self):
        arglist = [
            '--remote-id',
            self.idp.remote_ids[0],
            '--remote-id',
            self.idp.remote_ids[1],
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider_id', self.idp.id),
            ('remote_ids', self.idp.remote_ids),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'remote_ids': self.idp.remote_ids,
            'is_enabled': True,
        }

        self.identity_sdk_client.create_identity_provider.assert_called_with(
            id=self.idp.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)

    def test_create_identity_provider_remote_ids_file(self):
        arglist = [
            '--remote-id-file',
            '/tmp/file_name',
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider_id', self.idp.id),
            ('remote_id_file', '/tmp/file_name'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        mocker = mock.Mock()
        mocker.return_value = "\n".join(self.idp.remote_ids)
        with mock.patch(
            "openstackclient.identity.v3.identity_provider."
            "utils.read_blob_file_contents",
            mocker,
        ):
            columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'remote_ids': self.idp.remote_ids,
            'is_enabled': True,
        }

        self.identity_sdk_client.create_identity_provider.assert_called_with(
            id=self.idp.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)

    def test_create_identity_provider_disabled(self):
        idp_disabled = sdk_fakes.generate_fake_resource(
            _identity_provider.IdentityProvider,
            domain_id=self.domain.id,
            remote_ids=['entity1', 'entity2'],
            is_enabled=False,
        )
        self.identity_sdk_client.create_identity_provider.return_value = (
            idp_disabled
        )
        self.cmd = identity_provider.CreateIdentityProvider(self.app, None)

        arglist = [
            '--disable',
            idp_disabled.id,
        ]
        verifylist = [
            ('identity_provider_id', idp_disabled.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': False,
        }

        self.identity_sdk_client.create_identity_provider.assert_called_with(
            id=idp_disabled.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        datalist = (
            idp_disabled.authorization_ttl,
            idp_disabled.description,
            self.domain.id,
            False,
            idp_disabled.id,
            format_columns.ListColumn(self.idp.remote_ids),
        )
        self.assertCountEqual(datalist, data)

    def test_create_identity_provider_domain_name(self):
        arglist = [
            '--domain',
            self.domain.name,
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider_id', self.idp.id),
            ('domain', self.domain.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'domain_id': self.domain.id,
            'is_enabled': True,
        }

        self.identity_sdk_client.create_identity_provider.assert_called_with(
            id=self.idp.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)

    def test_create_identity_provider_domain_id(self):
        arglist = [
            '--domain',
            self.domain.id,
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider_id', self.idp.id),
            ('domain', self.domain.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'domain_id': self.domain.id,
            'is_enabled': True,
        }

        self.identity_sdk_client.create_identity_provider.assert_called_with(
            id=self.idp.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)

    def test_create_identity_provider_authttl_positive(self):
        arglist = [
            '--authorization-ttl',
            '60',
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider_id', self.idp.id),
            ('authorization_ttl', 60),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
            'authorization_ttl': 60,
        }

        self.identity_sdk_client.create_identity_provider.assert_called_with(
            id=self.idp.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)

    def test_create_identity_provider_authttl_zero(self):
        arglist = [
            '--authorization-ttl',
            '0',
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider_id', self.idp.id),
            ('authorization_ttl', 0),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
            'authorization_ttl': 0,
        }

        self.identity_sdk_client.create_identity_provider.assert_called_with(
            id=self.idp.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)

    def test_create_identity_provider_authttl_negative(self):
        arglist = [
            '--authorization-ttl',
            '-60',
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider_id', self.idp.id),
            ('authorization_ttl', -60),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_create_identity_provider_authttl_not_int(self):
        arglist = [
            '--authorization-ttl',
            'spam',
            self.idp.id,
        ]
        verifylist = []
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


class TestIdentityProviderDelete(identity_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.idp = sdk_fakes.generate_fake_resource(
            _identity_provider.IdentityProvider
        )

        self.identity_sdk_client.delete_identity_provider.return_value = None
        self.cmd = identity_provider.DeleteIdentityProvider(self.app, None)

    def test_delete_identity_provider(self):
        arglist = [
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider', [self.idp.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_identity_provider.assert_called_with(
            self.idp.id,
        )
        self.assertIsNone(result)


class TestIdentityProviderList(identity_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.domain = sdk_fakes.generate_fake_resource(_domain.Domain)
        self.idp = sdk_fakes.generate_fake_resource(
            _identity_provider.IdentityProvider,
            domain_id=self.domain.id,
            remote_ids=['entity1', 'entity2'],
        )

        self.identity_sdk_client.identity_providers.return_value = [self.idp]

        # Get the command object to test
        self.cmd = identity_provider.ListIdentityProvider(self.app, None)

    def test_identity_provider_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.identity_providers.assert_called_with()

        collist = ('ID', 'Enabled', 'Domain ID', 'Description')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.idp.id,
                True,
                self.domain.id,
                self.idp.description,
            ),
        )
        self.assertCountEqual(datalist, tuple(data))

    def test_identity_provider_list_ID_option(self):
        arglist = ['--id', self.idp.id]
        verifylist = [('id', self.idp.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {'id': self.idp.id}
        self.identity_sdk_client.identity_providers.assert_called_with(
            **kwargs
        )

        collist = ('ID', 'Enabled', 'Domain ID', 'Description')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.idp.id,
                True,
                self.domain.id,
                self.idp.description,
            ),
        )
        self.assertCountEqual(datalist, tuple(data))

    def test_identity_provider_list_enabled_option(self):
        arglist = ['--enabled']
        verifylist = [('enabled', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {'is_enabled': True}
        self.identity_sdk_client.identity_providers.assert_called_with(
            **kwargs
        )

        collist = ('ID', 'Enabled', 'Domain ID', 'Description')
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.idp.id,
                True,
                self.domain.id,
                self.idp.description,
            ),
        )
        self.assertCountEqual(datalist, tuple(data))


class TestIdentityProviderSet(identity_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.domain = sdk_fakes.generate_fake_resource(_domain.Domain)
        self.idp = sdk_fakes.generate_fake_resource(
            _identity_provider.IdentityProvider,
            domain_id=self.domain.id,
            remote_ids=['entity1', 'entity2'],
        )

        self.cmd = identity_provider.SetIdentityProvider(self.app, None)

    def test_identity_provider_set_description(self):
        """Set Identity Provider's description."""

        def prepare(self):
            """Prepare fake return objects before the test is executed"""
            new_description = 'new desc'

            self.updated_idp = sdk_fakes.generate_fake_resource(
                _identity_provider.IdentityProvider,
                id=self.idp.id,
                description=new_description,
                domain_id=self.domain.id,
                remote_ids=self.idp.remote_ids,
            )
            self.identity_sdk_client.update_identity_provider.return_value = (
                self.updated_idp
            )

        prepare(self)
        arglist = ['--description', self.updated_idp.description, self.idp.id]
        verifylist = [
            ('identity_provider', self.idp.id),
            ('description', self.updated_idp.description),
            ('enable', False),
            ('disable', False),
            ('remote_ids', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.identity_sdk_client.update_identity_provider.assert_called_with(
            self.idp.id,
            description=self.updated_idp.description,
        )

    def test_identity_provider_disable(self):
        """Disable Identity Provider

        Set Identity Provider's ``enabled`` attribute to False.
        """

        def prepare(self):
            """Prepare fake return objects before the test is executed"""
            self.updated_idp = sdk_fakes.generate_fake_resource(
                _identity_provider.IdentityProvider,
                id=self.idp.id,
                enabled=False,
                domain_id=self.domain.id,
                remote_ids=self.idp.remote_ids,
            )
            self.identity_sdk_client.update_identity_provider.return_value = (
                self.updated_idp
            )

        prepare(self)
        arglist = [
            '--disable',
            self.idp.id,
            '--remote-id',
            self.idp.remote_ids[0],
            '--remote-id',
            self.idp.remote_ids[1],
        ]
        verifylist = [
            ('identity_provider', self.idp.id),
            ('description', None),
            ('enable', False),
            ('disable', True),
            ('remote_ids', self.idp.remote_ids),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.identity_sdk_client.update_identity_provider.assert_called_with(
            self.idp.id,
            is_enabled=False,
            remote_ids=self.idp.remote_ids,
        )

    def test_identity_provider_enable(self):
        """Enable Identity Provider.

        Set Identity Provider's ``enabled`` attribute to True.
        """

        def prepare(self):
            """Prepare fake return objects before the test is executed"""
            self.updated_idp = sdk_fakes.generate_fake_resource(
                _identity_provider.IdentityProvider,
                id=self.idp.id,
                is_enabled=True,
                domain_id=self.domain.id,
                remote_ids=self.idp.remote_ids,
            )
            self.identity_sdk_client.update_identity_provider.return_value = (
                self.updated_idp
            )

        prepare(self)
        arglist = [
            '--enable',
            self.idp.id,
            '--remote-id',
            self.idp.remote_ids[0],
            '--remote-id',
            self.idp.remote_ids[1],
        ]
        verifylist = [
            ('identity_provider', self.idp.id),
            ('description', None),
            ('enable', True),
            ('disable', False),
            ('remote_ids', self.idp.remote_ids),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.identity_sdk_client.update_identity_provider.assert_called_with(
            self.idp.id,
            is_enabled=True,
            remote_ids=self.idp.remote_ids,
        )

    def test_identity_provider_replace_remote_ids(self):
        """Enable Identity Provider.

        Set Identity Provider's ``enabled`` attribute to True.
        """

        def prepare(self):
            """Prepare fake return objects before the test is executed"""
            self.new_remote_id = 'new_entity'

            self.updated_idp = sdk_fakes.generate_fake_resource(
                _identity_provider.IdentityProvider,
                id=self.idp.id,
                domain_id=self.domain.id,
                remote_ids=[self.new_remote_id],
            )
            self.identity_sdk_client.update_identity_provider.return_value = (
                self.updated_idp
            )

        prepare(self)
        arglist = [
            '--enable',
            self.idp.id,
            '--remote-id',
            self.new_remote_id,
        ]
        verifylist = [
            ('identity_provider', self.idp.id),
            ('description', None),
            ('enable', True),
            ('disable', False),
            ('remote_ids', [self.new_remote_id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.identity_sdk_client.update_identity_provider.assert_called_with(
            self.idp.id,
            is_enabled=True,
            remote_ids=[self.new_remote_id],
        )

    def test_identity_provider_replace_remote_ids_file(self):
        """Enable Identity Provider.

        Set Identity Provider's ``enabled`` attribute to True.
        """

        def prepare(self):
            """Prepare fake return objects before the test is executed"""
            self.new_remote_id = 'new_entity'

            self.updated_idp = sdk_fakes.generate_fake_resource(
                _identity_provider.IdentityProvider,
                id=self.idp.id,
                domain_id=self.domain.id,
                remote_ids=[self.new_remote_id],
            )
            self.identity_sdk_client.update_identity_provider.return_value = (
                self.updated_idp
            )

        prepare(self)
        arglist = [
            '--enable',
            self.idp.id,
            '--remote-id-file',
            self.new_remote_id,
        ]
        verifylist = [
            ('identity_provider', self.idp.id),
            ('description', None),
            ('enable', True),
            ('disable', False),
            ('remote_id_file', self.new_remote_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        mocker = mock.Mock()
        mocker.return_value = self.new_remote_id
        with mock.patch(
            "openstackclient.identity.v3.identity_provider."
            "utils.read_blob_file_contents",
            mocker,
        ):
            self.cmd.take_action(parsed_args)
        self.identity_sdk_client.update_identity_provider.assert_called_with(
            self.idp.id,
            is_enabled=True,
            remote_ids=[self.new_remote_id],
        )

    def test_identity_provider_no_options(self):
        def prepare(self):
            """Prepare fake return objects before the test is executed"""
            self.updated_idp = sdk_fakes.generate_fake_resource(
                _identity_provider.IdentityProvider,
                id=self.idp.id,
                domain_id=self.domain.id,
                remote_ids=self.idp.remote_ids,
            )
            self.identity_sdk_client.update_identity_provider.return_value = (
                self.updated_idp
            )

        prepare(self)
        arglist = [
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider', self.idp.id),
            ('enable', False),
            ('disable', False),
            ('remote_ids', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

    def test_identity_provider_set_authttl_positive(self):
        def prepare(self):
            """Prepare fake return objects before the test is executed"""
            self.updated_idp = sdk_fakes.generate_fake_resource(
                _identity_provider.IdentityProvider,
                id=self.idp.id,
                authorization_ttl=60,
                domain_id=self.domain.id,
                remote_ids=self.idp.remote_ids,
            )
            self.identity_sdk_client.update_identity_provider.return_value = (
                self.updated_idp
            )

        prepare(self)
        arglist = ['--authorization-ttl', '60', self.idp.id]
        verifylist = [
            ('identity_provider', self.idp.id),
            ('enable', False),
            ('disable', False),
            ('remote_ids', None),
            ('authorization_ttl', 60),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.identity_sdk_client.update_identity_provider.assert_called_with(
            self.idp.id,
            authorization_ttl=60,
        )

    def test_identity_provider_set_authttl_zero(self):
        def prepare(self):
            """Prepare fake return objects before the test is executed"""
            self.updated_idp = sdk_fakes.generate_fake_resource(
                _identity_provider.IdentityProvider,
                id=self.idp.id,
                authorization_ttl=0,
                domain_id=self.domain.id,
                remote_ids=self.idp.remote_ids,
            )
            self.identity_sdk_client.update_identity_provider.return_value = (
                self.updated_idp
            )

        prepare(self)
        arglist = ['--authorization-ttl', '0', self.idp.id]
        verifylist = [
            ('identity_provider', self.idp.id),
            ('enable', False),
            ('disable', False),
            ('remote_ids', None),
            ('authorization_ttl', 0),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.identity_sdk_client.update_identity_provider.assert_called_with(
            self.idp.id,
            authorization_ttl=0,
        )

    def test_identity_provider_set_authttl_negative(self):
        arglist = ['--authorization-ttl', '-1', self.idp.id]
        verifylist = [
            ('identity_provider', self.idp.id),
            ('enable', False),
            ('disable', False),
            ('remote_ids', None),
            ('authorization_ttl', -1),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_identity_provider_set_authttl_not_int(self):
        arglist = ['--authorization-ttl', 'spam', self.idp.id]
        verifylist = []
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


class TestIdentityProviderShow(identity_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.idp = sdk_fakes.generate_fake_resource(
            _identity_provider.IdentityProvider,
            remote_ids=['entity1', 'entity2'],
        )

        self.identity_sdk_client.get_identity_provider.return_value = self.idp

        self.cmd = identity_provider.ShowIdentityProvider(self.app, None)

    def test_identity_provider_show(self):
        arglist = [
            self.idp.id,
        ]
        verifylist = [
            ('identity_provider', self.idp.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.get_identity_provider.assert_called_with(
            self.idp.id,
        )

        collist = (
            'authorization_ttl',
            'description',
            'domain_id',
            'enabled',
            'id',
            'remote_ids',
        )
        self.assertEqual(collist, columns)
        datalist = (
            self.idp.authorization_ttl,
            self.idp.description,
            self.idp.domain_id,
            True,
            self.idp.id,
            format_columns.ListColumn(self.idp.remote_ids),
        )
        self.assertCountEqual(datalist, data)



# Copyright 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from collections.abc import Iterable, Sequence
import json
import logging
from typing import Any

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


def _format_property(prop: Any) -> dict[str, Any]:
    prop = prop.to_dict(ignore_none=True, original_names=True)
    return {
        key: prop[key]
        for key in [
            'namespace_name',
            'name',
            'type',
            'title',
            'description',
            'operators',
            'default',
            'is_readonly',
            'minimum',
            'maximum',
            'enum',
            'pattern',
            'min_length',
            'max_length',
            'items',
            'require_unique_items',
            'min_items',
            'max_items',
            'allow_additional_items',
        ]
        if key in prop
    }


class CreateMetadefProperty(command.ShowOne):
    _description = _("Create a metadef property")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--name",
            required=True,
            help=_("Internal name of the property"),
        )
        parser.add_argument(
            "--title",
            required=True,
            help=_("Property name displayed to the user"),
        )
        parser.add_argument(
            "--type",
            required=True,
            help=_("Property type"),
        )
        parser.add_argument(
            "--schema",
            required=True,
            help=_("Valid JSON schema of the property"),
        )
        parser.add_argument(
            "namespace",
            help=_("Name of namespace the property will belong."),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        image_client = self.app.client_manager.image

        kwargs = {
            'name': parsed_args.name,
            'title': parsed_args.title,
            'type': parsed_args.type,
        }
        try:
            kwargs.update(json.loads(parsed_args.schema))
        except json.JSONDecodeError as e:
            raise exceptions.CommandError(
                _("Failed to load JSON schema: %(e)s")
                % {
                    'e': e,
                }
            )

        data = image_client.create_metadef_property(
            parsed_args.namespace, **kwargs
        )
        info = _format_property(data)

        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data


class DeleteMetadefProperty(command.Command):
    _description = _("Delete metadef propert(ies)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_("Metadef namespace of the property (name)"),
        )
        parser.add_argument(
            "properties",
            metavar="<property>",
            nargs="*",
            help=_(
                "Metadef properties to delete (name) "
                "(omit this argument to delete all properties "
                "in the namespace)"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        image_client = self.app.client_manager.image

        if not parsed_args.properties:
            image_client.delete_all_metadef_properties(parsed_args.namespace)
            return

        result = 0
        for prop in parsed_args.properties:
            try:
                image_client.delete_metadef_property(
                    prop,
                    parsed_args.namespace,
                    ignore_missing=False,
                )
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete property with name or ID "
                        "'%(property)s' from namespace '%(namespace)s': %(e)s"
                    ),
                    {
                        'property': prop,
                        'namespace': parsed_args.namespace,
                        'e': e,
                    },
                )

        if result > 0:
            total = len(parsed_args.namespace)
            msg = _("%(result)s of %(total)s properties failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListMetadefProperties(command.Lister):
    _description = _("List metadef properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_("An identifier (a name) for the namespace"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        image_client = self.app.client_manager.image
        props = image_client.metadef_properties(parsed_args.namespace)
        columns = ['name', 'title', 'type']
        return (
            columns,
            (
                utils.get_item_properties(
                    prop,
                    columns,
                )
                for prop in props
            ),
        )


class SetMetadefProperty(command.Command):
    _description = _("Update metadef namespace property")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "--name",
            help=_("Internal name of the property"),
        )
        parser.add_argument(
            "--title",
            help=_("Property name displayed to the user"),
        )
        parser.add_argument(
            "--type",
            help=_("Property type"),
        )
        parser.add_argument(
            "--schema",
            help=_("Valid JSON schema of the property"),
        )
        parser.add_argument(
            "namespace",
            help=_("Namespace of the namespace to which the property belongs"),
        )
        parser.add_argument(
            "property",
            help=_("Property to update"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        image_client = self.app.client_manager.image

        # We need to pass the values for *all* attributes as kwargs to
        # update_metadef_property(), otherwise the attributes that are not
        # listed will be reset.
        data = image_client.get_metadef_property(
            parsed_args.property,
            parsed_args.namespace,
        )
        kwargs = _format_property(data)
        for key in ['name', 'title', 'type']:
            argument = getattr(parsed_args, key, None)
            if argument is not None:
                kwargs[key] = argument

        if parsed_args.schema:
            try:
                kwargs.update(json.loads(parsed_args.schema))
            except json.JSONDecodeError as e:
                raise exceptions.CommandError(
                    _("Failed to load JSON schema: %(e)s")
                    % {
                        'e': e,
                    }
                )

        image_client.update_metadef_property(
            parsed_args.property,
            parsed_args.namespace,
            **kwargs,
        )


class ShowMetadefProperty(command.ShowOne):
    _description = _("Show a particular metadef property")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "namespace",
            metavar="<namespace>",
            help=_("Metadef namespace of the property (name)"),
        )
        parser.add_argument(
            "property",
            metavar="<property>",
            help=_("Property to show"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        image_client = self.app.client_manager.image
        data = image_client.get_metadef_property(
            parsed_args.property,
            parsed_args.namespace,
        )
        info = _format_property(data)

        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Endpoint action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _
from openstackclient.identity import common


LOG = logging.getLogger(__name__)


class CreateEndpoint(command.ShowOne):
    _description = _("Create new endpoint")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'service',
            metavar='<service>',
            help=_('Service to be associated with new endpoint (name or ID)'),
        )
        parser.add_argument(
            '--publicurl',
            metavar='<url>',
            required=True,
            help=_('New endpoint public URL (required)'),
        )
        parser.add_argument(
            '--adminurl',
            metavar='<url>',
            help=_('New endpoint admin URL'),
        )
        parser.add_argument(
            '--internalurl',
            metavar='<url>',
            help=_('New endpoint internal URL'),
        )
        parser.add_argument(
            '--region',
            metavar='<region-id>',
            help=_('New endpoint region ID'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = self.app.client_manager.identity
        service = common.find_service(identity_client, parsed_args.service)
        endpoint = identity_client.endpoints.create(
            parsed_args.region,
            service.id,
            parsed_args.publicurl,
            parsed_args.adminurl,
            parsed_args.internalurl,
        )

        info = {}
        info.update(endpoint._info)
        info['service_name'] = service.name
        info['service_type'] = service.type
        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data


class DeleteEndpoint(command.Command):
    _description = _("Delete endpoint(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'endpoints',
            metavar='<endpoint-id>',
            nargs='+',
            help=_('Endpoint(s) to delete (ID only)'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = self.app.client_manager.identity

        result = 0
        for endpoint in parsed_args.endpoints:
            try:
                identity_client.endpoints.delete(endpoint)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete endpoint with "
                        "ID '%(endpoint)s': %(e)s"
                    ),
                    {'endpoint': endpoint, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.endpoints)
            msg = _("%(result)s of %(total)s endpoints failed to delete.") % {
                'result': result,
                'total': total,
            }
            raise exceptions.CommandError(msg)


class ListEndpoint(command.Lister):
    _description = _("List endpoints")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        identity_client = self.app.client_manager.identity

        columns: tuple[str, ...] = (
            'ID',
            'Region',
            'Service Name',
            'Service Type',
        )
        if parsed_args.long:
            columns += (
                'PublicURL',
                'AdminURL',
                'InternalURL',
            )
        data = identity_client.endpoints.list()

        for ep in data:
            service = common.find_service(identity_client, ep.service_id)
            ep.service_name = service.name
            ep.service_type = service.type
        return (
            columns,
            (
                utils.get_item_properties(
                    s,
                    columns,
                    formatters={},
                )
                for s in data
            ),
        )


class ShowEndpoint(command.ShowOne):
    _description = _("Display endpoint details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'endpoint_or_service',
            metavar='<endpoint>',
            help=_(
                'Endpoint to display (endpoint ID, service ID,'
                ' service name, service type)'
            ),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        identity_client = self.app.client_manager.identity
        data = identity_client.endpoints.list()
        match = None
        for ep in data:
            if ep.id == parsed_args.endpoint_or_service:
                match = ep
                service = common.find_service(identity_client, ep.service_id)
        if match is None:
            service = common.find_service(
                identity_client, parsed_args.endpoint_or_service
            )
            for ep in data:
                if ep.service_id == service.id:
                    match = ep
        if match is None:
            return ((), ())

        info = {}
        info.update(match._info)
        info['service_name'] = service.name
        info['service_type'] = service.type
        col_headers, col_data = zip(*sorted(info.items()))
        return col_headers, col_data



# Copyright 2016 FUJITSU LIMITED
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import re
from unittest import mock

from openstack.network.v2 import firewall_policy
from openstack.network.v2 import firewall_rule
from openstack.test import fakes as sdk_fakes
from osc_lib import exceptions

from openstackclient.network.v2.fwaas import policy as fwaas_policy
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as test_utils

CONVERT_MAP = {
    'project': 'project_id',
    'port': 'ports',
    'name': 'name',
    'id': 'id',
    'firewall_rule': 'firewall_rules',
    'description': 'description',
}


def _generate_data(source=None, data=None):
    source = source or {}
    if data:
        source.update(data)
    return tuple(source[key] for key in source)


def _generate_req_and_res(verifylist, response):
    request = dict(verifylist)
    for key, val in verifylist:
        converted = CONVERT_MAP.get(key, key)
        del request[key]
        if re.match('^no_', key) and val is True:
            new_value = None
        elif val is True or val is False:
            new_value = val
        elif key in ('name', 'description'):
            new_value = val
        else:
            new_value = val
        request[converted] = new_value
        response[converted] = new_value
    return request, response


class TestFirewallPolicy(network_fakes.TestNetworkV2):
    def check_results(self, headers, data, exp_req, is_list=False):
        if is_list:
            req_body = {self.res_plural: [exp_req]}
        else:
            req_body = exp_req
        self.mocked.assert_called_once_with(**req_body)
        self.assertEqual(self.ordered_headers, tuple(sorted(headers)))

    def setUp(self):
        super().setUp()

        self.identity_client.projects.get.side_effect = lambda x: mock.Mock(
            id=x
        )
        self.res = 'firewall_policy'
        self.res_plural = 'firewall_policies'
        self.resource = sdk_fakes.generate_fake_resource(
            firewall_policy.FirewallPolicy
        )
        # TODO(slaweq): Remove this "firewall_rules" override once bug
        # https://bugs.launchpad.net/openstacksdk/+bug/2146537 will be fixed in
        # OpenStackSDK
        self.resource['firewall_rules'] = [self.resource['firewall_rules']]

        self.list_headers = (
            'ID',
            'Name',
            'Firewall Rules',
        )
        self.list_data = (
            self.resource['id'],
            self.resource['name'],
            self.resource['firewall_rules'],
        )
        self.headers = tuple(
            (*self.list_headers, 'Description', 'Audited', 'Shared', 'Project')
        )
        self.data = _generate_data(self.resource)
        self.ordered_headers = (
            'Audited',
            'Description',
            'Firewall Rules',
            'ID',
            'Name',
            'Project',
            'Shared',
        )
        self.ordered_data = (
            self.resource['audited'],
            self.resource['description'],
            self.resource['firewall_rules'],
            self.resource['id'],
            self.resource['name'],
            self.resource['project_id'],
            self.resource['shared'],
        )
        self.ordered_columns = (
            'audited',
            'description',
            'firewall_rules',
            'id',
            'name',
            'project_id',
            'shared',
        )


class TestCreateFirewallPolicy(TestFirewallPolicy):
    def setUp(self):
        super().setUp()
        self.network_client.create_firewall_policy.return_value = self.resource
        self.mocked = self.network_client.create_firewall_policy
        self.cmd = fwaas_policy.CreateFirewallPolicy(self.app, None)

    def _update_expect_response(self, request, response):
        """Set expected request and response

        :param request
            A dictionary of request body(dict of verifylist)
        :param response
            A OrderedDict of request body
        """
        # Update response body
        pass  # identity_client.projects.get is already mocked in setUp
        # Update response(finally returns 'data')
        self.data = _generate_data(data=response)
        self.ordered_data = tuple(
            response[column] for column in self.ordered_columns
        )

    def test_create_with_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_with_mandatory_param(self):
        name = 'my-fwg'
        arglist = [
            name,
        ]
        verifylist = [
            ('name', name),
        ]
        request, response = _generate_req_and_res(verifylist, self.resource)
        self._update_expect_response(request, response)
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.check_results(headers, data, request)

    def test_create_with_rules(self):
        name = 'my-fwg'
        rule1 = 'rule1'
        rule2 = 'rule2'

        def _mock_find_rule(*args, **kwargs):
            return sdk_fakes.generate_fake_resource(
                firewall_rule.FirewallRule, id=args[0]
            )

        self.network_client.find_firewall_rule.side_effect = _mock_find_rule

        arglist = [
            name,
            '--firewall-rule',
            rule1,
            '--firewall-rule',
            rule2,
        ]
        verifylist = [
            ('name', name),
            ('firewall_rule', [rule1, rule2]),
        ]
        request, response = _generate_req_and_res(verifylist, self.resource)
        self._update_expect_response(request, response)
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)
        self.assertEqual(2, self.network_client.find_firewall_rule.call_count)

        self.check_results(headers, data, request)

    def test_create_with_all_params(self):
        name = 'my-fwp'
        desc = 'my-desc'
        rule1 = 'rule1'
        rule2 = 'rule2'
        project = 'my-tenant'

        def _mock_find(*args, **kwargs):
            if self.res in args[0]:
                rules = self.resource['firewall_rules']
                return mock.Mock(id=args[0], firewall_rules=rules)
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_policy.side_effect = _mock_find
        self.network_client.find_firewall_rule.side_effect = _mock_find

        arglist = [
            name,
            '--description',
            desc,
            '--firewall-rule',
            rule1,
            '--firewall-rule',
            rule2,
            '--project',
            project,
            '--share',
            '--audited',
        ]
        verifylist = [
            ('name', name),
            ('description', desc),
            ('firewall_rule', [rule1, rule2]),
            ('project', project),
            ('shared', True),
            ('audited', True),
        ]
        request, response = _generate_req_and_res(verifylist, self.resource)
        self._update_expect_response(request, response)
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.check_results(headers, data, request)

    def test_create_with_firewall_rule_and_no(self):
        name = 'my-fwp'
        rule1 = 'rule1'
        rule2 = 'rule2'
        arglist = [
            name,
            '--firewall-rule',
            rule1,
            '--firewall-rule',
            rule2,
            '--no-firewall-rule',
        ]
        verifylist = [
            ('name', name),
            ('firewall_rule', [rule1, rule2]),
            ('no_firewall_rule', True),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_with_shared_and_no_share(self):
        name = 'my-fwp'
        arglist = [
            name,
            '--share',
            '--no-share',
        ]
        verifylist = [
            ('name', name),
            ('shared', True),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_with_audited_and_no(self):
        name = 'my-fwp'
        arglist = [
            name,
            '--audited',
            '--no-audited',
        ]
        verifylist = [
            ('name', name),
            ('audited', True),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


class TestListFirewallPolicy(TestFirewallPolicy):
    def setUp(self):
        super().setUp()
        self.network_client.firewall_policies.return_value = [self.resource]
        self.mocked = self.network_client.firewall_policies
        self.cmd = fwaas_policy.ListFirewallPolicy(self.app, None)

    def test_list_with_no_option(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with()
        self.assertEqual(list(self.list_headers), headers)
        self.assertEqual([self.list_data], list(data))

    def test_list_with_long_option(self):
        arglist = ['--long']
        verifylist = [('long', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, _data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with()
        self.assertEqual(list(self.headers), headers)


class TestShowFirewallPolicy(TestFirewallPolicy):
    def setUp(self):
        super().setUp()
        self.network_client.get_firewall_policy.return_value = self.resource
        self.mocked = self.network_client.get_firewall_policy
        self.cmd = fwaas_policy.ShowFirewallPolicy(self.app, None)

    def test_show_filtered_by_id_or_name(self):
        target = self.resource['id']

        def _mock_fwaas(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_policy.side_effect = _mock_fwaas

        arglist = [target]
        verifylist = [(self.res, target)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, _data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target)
        self.assertEqual(self.ordered_headers, headers)


class TestSetFirewallPolicy(TestFirewallPolicy):
    def setUp(self):
        super().setUp()
        self.network_client.update_firewall_policy.return_value = self.resource
        self.mocked = self.network_client.update_firewall_policy

        def _mock_find_rule(*args, **kwargs):
            return sdk_fakes.generate_fake_resource(
                firewall_rule.FirewallRule, id=args[0]
            )

        def _mock_find_policy(*args, **kwargs):
            return self.resource

        self.network_client.find_firewall_policy.side_effect = (
            _mock_find_policy
        )
        self.network_client.find_firewall_rule.side_effect = _mock_find_rule

        self.cmd = fwaas_policy.SetFirewallPolicy(self.app, None)

    def test_set_name(self):
        target = self.resource['id']
        update = 'change'
        arglist = [target, '--name', update]
        verifylist = [
            (self.res, target),
            ('name', update),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target, **{'name': update})
        self.assertIsNone(result)

    def test_set_options(self):
        target = self.resource['id']
        updated_desc = 'change-desc'
        arglist = [target, '--description', updated_desc, '--share']
        verifylist = [
            (self.res, target),
            ('description', updated_desc),
            ('shared', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(
            target, shared=True, description=updated_desc
        )
        self.assertIsNone(result)

    def test_set_rules(self):
        target = self.resource['id']
        rule1 = 'new_rule1'
        rule2 = 'new_rule2'
        arglist = [
            target,
            '--firewall-rule',
            rule1,
            '--firewall-rule',
            rule2,
        ]
        verifylist = [
            (self.res, target),
            ('firewall_rule', [rule1, rule2]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        expect = self.resource['firewall_rules'] + [rule1, rule2]
        body = {'firewall_rules': expect}
        self.mocked.assert_called_once_with(target, **body)
        self.assertEqual(2, self.network_client.find_firewall_rule.call_count)
        self.assertEqual(
            2, self.network_client.find_firewall_policy.call_count
        )
        self.assertIsNone(result)

    def test_set_no_rules(self):
        target = self.resource['id']
        arglist = [target, '--no-firewall-rule']
        verifylist = [
            (self.res, target),
            ('no_firewall_rule', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        body = {'firewall_rules': []}
        self.mocked.assert_called_once_with(target, **body)
        self.assertIsNone(result)

    def test_set_rules_and_no_rules(self):
        target = self.resource['id']
        rule1 = 'rule1'
        arglist = [
            target,
            '--firewall-rule',
            rule1,
            '--no-firewall-rule',
        ]
        verifylist = [
            (self.res, target),
            ('firewall_rule', [rule1]),
            ('no_firewall_rule', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        body = {'firewall_rules': [rule1]}
        self.mocked.assert_called_once_with(target, **body)
        self.assertEqual(1, self.network_client.find_firewall_rule.call_count)
        self.assertEqual(
            1, self.network_client.find_firewall_policy.call_count
        )
        self.assertIsNone(result)

    def test_set_audited(self):
        target = self.resource['id']
        arglist = [target, '--audited']
        verifylist = [
            (self.res, target),
            ('audited', True),
        ]
        body = {'audited': True}
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target, **body)
        self.assertIsNone(result)

    def test_set_no_audited(self):
        target = self.resource['id']
        arglist = [target, '--no-audited']
        verifylist = [
            (self.res, target),
            ('audited', False),
        ]
        body = {'audited': False}
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target, **body)
        self.assertIsNone(result)

    def test_set_audited_and_no_audited(self):
        target = self.resource['id']
        arglist = [
            target,
            '--audited',
            '--no-audited',
        ]
        verifylist = [
            (self.res, target),
            ('audited', True),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_set_and_raises(self):
        self.network_client.update_firewall_policy.side_effect = Exception
        target = self.resource['id']

        arglist = [target, '--name', 'my-name']
        verifylist = [(self.res, target), ('name', 'my-name')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestDeleteFirewallPolicy(TestFirewallPolicy):
    def setUp(self):
        super().setUp()
        self.network_client.delete_firewall_policy.return_value = {
            self.res: self.resource
        }
        self.mocked = self.network_client.delete_firewall_policy
        self.cmd = fwaas_policy.DeleteFirewallPolicy(self.app, None)

    def test_delete_with_one_resource(self):
        target = self.resource['id']

        def _mock_fwaas(*args, **kwargs):
            return self.resource

        self.network_client.find_firewall_policy.side_effect = _mock_fwaas

        arglist = [target]
        verifylist = [(self.res, [target])]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(target)
        self.assertIsNone(result)

    def test_delete_with_multiple_resources(self):

        def _mock_fwaas(*args, **kwargs):
            return sdk_fakes.generate_fake_resource(
                firewall_policy.FirewallPolicy, id=args[0]
            )

        self.network_client.find_firewall_policy.side_effect = _mock_fwaas

        target1 = 'target1'
        target2 = 'target2'
        arglist = [target1, target2]
        verifylist = [(self.res, [target1, target2])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        self.assertIsNone(result)

        self.assertEqual(2, self.mocked.call_count)
        for idx, reference in enumerate([target1, target2]):
            actual = ''.join(self.mocked.call_args_list[idx][0][0])
            self.assertEqual(reference, actual)

    def test_delete_multiple_with_exception(self):
        target1 = 'target1'
        target2 = 'target2'
        arglist = [target1, target2]
        verifylist = [(self.res, [target1, target2])]

        def _mock_find(*args, **kwargs):
            if args[0] == target2:
                raise Exception('Not found')
            return self.resource

        self.network_client.find_firewall_policy.side_effect = _mock_find

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestFirewallPolicyInsertRule(TestFirewallPolicy):
    def setUp(self):
        super().setUp()
        self.network_client.insert_rule_into_policy.return_value = {
            self.res: self.resource
        }
        self.mocked = self.network_client.insert_rule_into_policy

        def _mock_find_policy(*args, **kwargs):
            return self.resource

        def _mock_find_rule(*args, **kwargs):
            return sdk_fakes.generate_fake_resource(
                firewall_rule.FirewallRule, id=args[0]
            )

        self.network_client.find_firewall_policy.side_effect = (
            _mock_find_policy
        )
        self.network_client.find_firewall_rule.side_effect = _mock_find_rule

        self.cmd = fwaas_policy.FirewallPolicyInsertRule(self.app, None)

    def test_insert_firewall_rule(self):
        target = self.resource['id']
        rule = 'new-rule'
        before = 'before'
        after = 'after'
        arglist = [
            target,
            rule,
            '--insert-before',
            before,
            '--insert-after',
            after,
        ]
        verifylist = [
            (self.res, target),
            ('firewall_rule', rule),
            ('insert_before', before),
            ('insert_after', after),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        body = {
            'firewall_rule_id': rule,
            'insert_before': before,
            'insert_after': after,
        }
        self.mocked.assert_called_once_with(target, **body)
        self.assertIsNone(result)
        self.assertEqual(
            1, self.network_client.find_firewall_policy.call_count
        )
        self.assertEqual(3, self.network_client.find_firewall_rule.call_count)

    def test_insert_with_no_firewall_rule(self):
        target = self.resource['id']
        arglist = [
            target,
        ]
        verifylist = [
            (self.res, target),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


class TestFirewallPolicyRemoveRule(TestFirewallPolicy):
    def setUp(self):
        super().setUp()
        self.network_client.remove_rule_from_policy.return_value = {
            self.res: self.resource
        }
        self.mocked = self.network_client.remove_rule_from_policy

        def _mock_find_policy(*args, **kwargs):
            return mock.Mock(id=args[0])

        self.network_client.find_firewall_policy.side_effect = (
            _mock_find_policy
        )
        self.network_client.find_firewall_rule.side_effect = _mock_find_policy

        self.cmd = fwaas_policy.FirewallPolicyRemoveRule(self.app, None)

    def test_remove_firewall_rule(self):
        target = self.resource['id']
        rule = 'remove-rule'
        arglist = [
            target,
            rule,
        ]
        verifylist = [
            (self.res, target),
            ('firewall_rule', rule),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        body = {'firewall_rule_id': rule}
        self.mocked.assert_called_once_with(target, **body)
        self.assertIsNone(result)
        self.assertEqual(
            1, self.network_client.find_firewall_policy.call_count
        )
        self.assertEqual(1, self.network_client.find_firewall_rule.call_count)

    def test_remove_with_no_firewall_rule(self):
        target = self.resource['id']
        arglist = [
            target,
        ]
        verifylist = [
            (self.res, target),
        ]
        self.assertRaises(
            test_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )


class TestUnsetFirewallPolicy(TestFirewallPolicy):
    def setUp(self):
        super().setUp()
        self.network_client.update_firewall_policy.return_value = {
            self.res: self.resource
        }
        self.mocked = self.network_client.update_firewall_policy

        def _mock_find_rule(*args, **kwargs):
            return sdk_fakes.generate_fake_resource(
                firewall_rule.FirewallRule, id=args[0]
            )

        def _mock_find_policy(*args, **kwargs):
            return mock.Mock(
                id=args[0],
                firewall_rules=self.resource['firewall_rules'],
            )

        self.network_client.find_firewall_policy.side_effect = (
            _mock_find_policy
        )
        self.network_client.find_firewall_rule.side_effect = _mock_find_rule

        self.cmd = fwaas_policy.UnsetFirewallPolicy(self.app, None)

    def test_unset_shared(self):
        target = self.resource['id']
        arglist = [
            target,
            '--share',
        ]
        verifylist = [
            (self.res, target),
            ('share', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        with mock.patch.object(fwaas_policy.LOG, 'warning') as mock_warning:
            result = self.cmd.take_action(parsed_args)
            mock_warning.assert_called_once_with(
                'The --share option is deprecated, please use '
                '"firewall policy set --no-share" instead.'
            )
        self.mocked.assert_called_once_with(target, **{'shared': False})
        self.assertIsNone(result)

    def test_unset_audited(self):
        target = self.resource['id']
        arglist = [
            target,
            '--audited',
        ]
        verifylist = [
            (self.res, target),
            ('audited', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        body = {'audited': False}
        self.mocked.assert_called_once_with(target, **body)
        self.assertIsNone(result)

    def test_unset_firewall_rule_not_matched(self):
        self.resource['firewall_rules'] = ['old_rule']
        target = self.resource['id']
        rule = 'new_rule'
        arglist = [
            target,
            '--firewall-rule',
            rule,
        ]
        verifylist = [
            (self.res, target),
            ('firewall_rule', [rule]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        body = {'firewall_rules': self.resource['firewall_rules']}
        self.mocked.assert_called_once_with(target, **body)
        self.assertIsNone(result)

    def test_unset_firewall_rule_matched(self):
        self.resource['firewall_rules'] = ['rule1', 'rule2']
        target = self.resource['id']
        rule = 'rule1'
        arglist = [
            target,
            '--firewall-rule',
            rule,
        ]
        verifylist = [
            (self.res, target),
            ('firewall_rule', [rule]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        body = {'firewall_rules': ['rule2']}
        self.mocked.assert_called_once_with(target, **body)
        self.assertIsNone(result)
        self.assertEqual(
            2, self.network_client.find_firewall_policy.call_count
        )
        self.assertEqual(1, self.network_client.find_firewall_rule.call_count)

    def test_unset_all_firewall_rule(self):
        target = self.resource['id']
        arglist = [
            target,
            '--all-firewall-rule',
        ]
        verifylist = [
            (self.res, target),
            ('all_firewall_rule', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        body = {'firewall_rules': []}
        self.mocked.assert_called_once_with(target, **body)
        self.assertIsNone(result)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import random
from unittest.mock import call

from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.network.v2 import network
from openstackclient.tests.unit.identity.v2_0 import fakes as identity_fakes_v2
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes_v3
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


# Tests for Neutron network
#
class TestNetwork(network_fakes.TestNetworkV2):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.identity_client.projects
        # Get a shortcut to the DomainManager Mock
        self.domains_mock = self.identity_client.domains


class TestCreateNetworkIdentityV3(TestNetwork):
    project = identity_fakes_v3.FakeProject.create_one_project()
    domain = identity_fakes_v3.FakeDomain.create_one_domain()
    # The new network created.
    _network = network_fakes.create_one_network(
        attrs={
            'project_id': project.id,
            'availability_zone_hints': ["nova"],
        }
    )
    qos_policy = network_fakes.create_one_qos_policy(
        attrs={'id': _network.qos_policy_id}
    )

    columns = (
        'admin_state_up',
        'availability_zone_hints',
        'availability_zones',
        'created_at',
        'description',
        'dns_domain',
        'id',
        'ipv4_address_scope',
        'ipv6_address_scope',
        'is_default',
        'is_vlan_transparent',
        'is_vlan_qinq',
        'mtu',
        'name',
        'port_security_enabled',
        'project_id',
        'provider:network_type',
        'provider:physical_network',
        'provider:segmentation_id',
        'qos_policy_id',
        'router:external',
        'shared',
        'status',
        'segments',
        'subnets',
        'tags',
        'revision_number',
        'updated_at',
    )

    data = (
        network.AdminStateColumn(_network.is_admin_state_up),
        format_columns.ListColumn(_network.availability_zone_hints),
        format_columns.ListColumn(_network.availability_zones),
        _network.created_at,
        _network.description,
        _network.dns_domain,
        _network.id,
        _network.ipv4_address_scope_id,
        _network.ipv6_address_scope_id,
        _network.is_default,
        _network.mtu,
        _network.name,
        _network.is_port_security_enabled,
        _network.project_id,
        _network.provider_network_type,
        _network.provider_physical_network,
        _network.provider_segmentation_id,
        _network.qos_policy_id,
        network.RouterExternalColumn(_network.is_router_external),
        _network.is_shared,
        _network.is_vlan_transparent,
        _network.is_vlan_qinq,
        _network.status,
        _network.segments,
        format_columns.ListColumn(_network.subnet_ids),
        format_columns.ListColumn(_network.tags),
        _network.revision_number,
        _network.updated_at,
    )

    def setUp(self):
        super().setUp()

        self.network_client.create_network.return_value = self._network

        self.network_client.set_tags.return_value = None

        # Get the command object to test
        self.cmd = network.CreateNetwork(self.app, None)

        self.projects_mock.get.return_value = self.project
        self.domains_mock.get.return_value = self.domain
        self.network_client.find_qos_policy.return_value = self.qos_policy

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_create_default_options(self):
        arglist = [
            self._network.name,
        ]
        verifylist = [
            ('name', self._network.name),
            ('enable', True),
            ('share', None),
            ('project', None),
            ('external', False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_network.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self._network.name,
            }
        )
        self.assertFalse(self.network_client.set_tags.called)
        self.assertEqual(set(self.columns), set(columns))
        self.assertCountEqual(self.data, data)

    def test_create_all_options(self):
        arglist = [
            "--disable",
            "--share",
            "--description",
            self._network.description,
            "--mtu",
            str(self._network.mtu),
            "--project",
            self.project.name,
            "--project-domain",
            self.domain.name,
            "--availability-zone-hint",
            "nova",
            "--external",
            "--default",
            "--provider-network-type",
            "vlan",
            "--provider-physical-network",
            "physnet1",
            "--provider-segment",
            "400",
            "--qos-policy",
            self.qos_policy.id,
            "--transparent-vlan",
            "--no-qinq-vlan",
            "--enable-port-security",
            "--dns-domain",
            "example.org.",
            self._network.name,
        ]
        verifylist = [
            ('disable', True),
            ('share', True),
            ('description', self._network.description),
            ('mtu', str(self._network.mtu)),
            ('project', self.project.name),
            ('project_domain', self.domain.name),
            ('availability_zone_hints', ["nova"]),
            ('external', True),
            ('default', True),
            ('provider_network_type', 'vlan'),
            ('physical_network', 'physnet1'),
            ('segmentation_id', '400'),
            ('qos_policy', self.qos_policy.id),
            ('transparent_vlan', True),
            ('qinq_vlan', False),
            ('enable_port_security', True),
            ('name', self._network.name),
            ('dns_domain', 'example.org.'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_network.assert_called_once_with(
            **{
                'admin_state_up': False,
                'availability_zone_hints': ["nova"],
                'name': self._network.name,
                'shared': True,
                'description': self._network.description,
                'mtu': str(self._network.mtu),
                'project_id': self.project.id,
                'is_default': True,
                'router:external': True,
                'provider:network_type': 'vlan',
                'provider:physical_network': 'physnet1',
                'provider:segmentation_id': '400',
                'qos_policy_id': self.qos_policy.id,
                'vlan_transparent': True,
                'vlan_qinq': False,
                'port_security_enabled': True,
                'dns_domain': 'example.org.',
            }
        )
        self.assertEqual(set(self.columns), set(columns))
        self.assertCountEqual(self.data, data)

    def test_create_other_options(self):
        arglist = [
            "--enable",
            "--no-share",
            "--disable-port-security",
            "--qinq-vlan",
            self._network.name,
        ]
        verifylist = [
            ('enable', True),
            ('no_share', True),
            ('name', self._network.name),
            ('external', False),
            ('qinq_vlan', True),
            ('disable_port_security', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_network.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self._network.name,
                'shared': False,
                'vlan_qinq': True,
                'port_security_enabled': False,
            }
        )
        self.assertEqual(set(self.columns), set(columns))
        self.assertCountEqual(self.data, data)

    def _test_create_with_tag(self, add_tags=True):
        arglist = [self._network.name]
        if add_tags:
            arglist += ['--tag', 'red', '--tag', 'blue']
        else:
            arglist += ['--no-tag']
        verifylist = [
            ('name', self._network.name),
            ('enable', True),
            ('share', None),
            ('project', None),
            ('external', False),
        ]
        if add_tags:
            verifylist.append(('tags', ['red', 'blue']))
        else:
            verifylist.append(('no_tag', True))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_network.assert_called_once_with(
            name=self._network.name, admin_state_up=True
        )
        if add_tags:
            self.network_client.set_tags.assert_called_once_with(
                self._network, tests_utils.CompareBySet(['red', 'blue'])
            )
        else:
            self.assertFalse(self.network_client.set_tags.called)
        self.assertEqual(set(self.columns), set(columns))
        self.assertCountEqual(self.data, data)

    def test_create_with_tags(self):
        self._test_create_with_tag(add_tags=True)

    def test_create_with_no_tag(self):
        self._test_create_with_tag(add_tags=False)

    def test_create_with_vlan_qinq_and_transparency_enabled(self):
        arglist = [
            "--transparent-vlan",
            "--qinq-vlan",
            self._network.name,
        ]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

    def test_create_with_provider_segment_without_provider_type(self):
        arglist = [
            "--provider-segment",
            "123",
            self._network.name,
        ]
        verifylist = [
            ('provider_network_type', None),
            ('segmentation_id', "123"),
            ('name', self._network.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )


class TestCreateNetworkIdentityV2(
    identity_fakes_v2.FakeClientMixin,
    network_fakes.FakeClientMixin,
    tests_utils.TestCommand,
):
    project = identity_fakes_v2.FakeProject.create_one_project()
    # The new network created.
    _network = network_fakes.create_one_network(
        attrs={'project_id': project.id}
    )
    columns = (
        'admin_state_up',
        'availability_zone_hints',
        'availability_zones',
        'created_at',
        'description',
        'dns_domain',
        'id',
        'ipv4_address_scope',
        'ipv6_address_scope',
        'is_default',
        'is_vlan_transparent',
        'is_vlan_qinq',
        'mtu',
        'name',
        'port_security_enabled',
        'project_id',
        'provider:network_type',
        'provider:physical_network',
        'provider:segmentation_id',
        'qos_policy_id',
        'router:external',
        'shared',
        'status',
        'segments',
        'subnets',
        'tags',
        'revision_number',
        'updated_at',
    )

    data = (
        network.AdminStateColumn(_network.is_admin_state_up),
        format_columns.ListColumn(_network.availability_zone_hints),
        format_columns.ListColumn(_network.availability_zones),
        _network.created_at,
        _network.description,
        _network.dns_domain,
        _network.id,
        _network.ipv4_address_scope_id,
        _network.ipv6_address_scope_id,
        _network.is_default,
        _network.mtu,
        _network.name,
        _network.is_port_security_enabled,
        _network.project_id,
        _network.provider_network_type,
        _network.provider_physical_network,
        _network.provider_segmentation_id,
        _network.qos_policy_id,
        network.RouterExternalColumn(_network.is_router_external),
        _network.is_shared,
        _network.is_vlan_transparent,
        _network.is_vlan_qinq,
        _network.status,
        _network.segments,
        format_columns.ListColumn(_network.subnet_ids),
        format_columns.ListColumn(_network.tags),
        _network.revision_number,
        _network.updated_at,
    )

    def setUp(self):
        super().setUp()

        self.network_client.create_network.return_value = self._network

        self.network_client.set_tags.return_value = None

        # Get the command object to test
        self.cmd = network.CreateNetwork(self.app, None)

        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.identity_client.tenants
        self.projects_mock.get.return_value = self.project

        # There is no DomainManager Mock in fake identity v2.

    def test_create_with_project_identityv2(self):
        arglist = [
            "--project",
            self.project.name,
            self._network.name,
        ]
        verifylist = [
            ('enable', True),
            ('share', None),
            ('name', self._network.name),
            ('project', self.project.name),
            ('external', False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.create_network.assert_called_once_with(
            **{
                'admin_state_up': True,
                'name': self._network.name,
                'project_id': self.project.id,
            }
        )
        self.assertFalse(self.network_client.set_tags.called)
        self.assertEqual(set(self.columns), set(columns))
        self.assertCountEqual(self.data, data)

    def test_create_with_domain_identityv2(self):
        arglist = [
            "--project",
            self.project.name,
            "--project-domain",
            "domain-name",
            self._network.name,
        ]
        verifylist = [
            ('enable', True),
            ('share', None),
            ('project', self.project.name),
            ('project_domain', "domain-name"),
            ('name', self._network.name),
            ('external', False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            AttributeError,
            self.cmd.take_action,
            parsed_args,
        )


class TestDeleteNetwork(TestNetwork):
    def setUp(self):
        super().setUp()

        # The networks to delete
        self._networks = network_fakes.create_networks(count=3)

        self.network_client.delete_network.return_value = None

        self.network_client.find_network = network_fakes.get_networks(
            networks=self._networks
        )

        # Get the command object to test
        self.cmd = network.DeleteNetwork(self.app, None)

    def test_delete_one_network(self):
        arglist = [
            self._networks[0].name,
        ]
        verifylist = [
            ('network', [self._networks[0].name]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_network.assert_called_once_with(
            self._networks[0]
        )
        self.assertIsNone(result)

    def test_delete_multiple_networks(self):
        arglist = []
        for n in self._networks:
            arglist.append(n.id)
        verifylist = [
            ('network', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for n in self._networks:
            calls.append(call(n))
        self.network_client.delete_network.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_multiple_networks_exception(self):
        arglist = [
            self._networks[0].id,
            'xxxx-yyyy-zzzz',
            self._networks[1].id,
        ]
        verifylist = [
            ('network', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Fake exception in find_network()
        ret_find = [
            self._networks[0],
            exceptions.NotFound('404'),
            self._networks[1],
        ]
        self.network_client.find_network.side_effect = ret_find

        # Fake exception in delete_network()
        ret_delete = [
            None,
            exceptions.NotFound('404'),
        ]
        self.network_client.delete_network.side_effect = ret_delete

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

        # The second call of find_network() should fail. So delete_network()
        # was only called twice.
        calls = [
            call(self._networks[0]),
            call(self._networks[1]),
        ]
        self.network_client.delete_network.assert_has_calls(calls)


class TestListNetwork(TestNetwork):
    # The networks going to be listed up.
    _network = network_fakes.create_networks(count=3)

    columns = (
        'ID',
        'Name',
        'Subnets',
    )
    columns_long = (
        'ID',
        'Name',
        'Status',
        'Project',
        'State',
        'Shared',
        'Subnets',
        'Network Type',
        'Router Type',
        'Availability Zones',
        'Tags',
    )

    data = []
    for net in _network:
        data.append(
            (
                net.id,
                net.name,
                format_columns.ListColumn(net.subnet_ids),
            )
        )

    data_long = []
    for net in _network:
        data_long.append(
            (
                net.id,
                net.name,
                net.status,
                net.project_id,
                network.AdminStateColumn(net.is_admin_state_up),
                net.is_shared,
                format_columns.ListColumn(net.subnet_ids),
                net.provider_network_type,
                network.RouterExternalColumn(net.is_router_external),
                format_columns.ListColumn(net.availability_zones),
                format_columns.ListColumn(net.tags),
            )
        )

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = network.ListNetwork(self.app, None)

        self.network_client.networks.return_value = self._network

        self._agent = network_fakes.create_one_network_agent()
        self.network_client.get_agent.return_value = self._agent

        self.network_client.dhcp_agent_hosting_networks.return_value = (
            self._network
        )

        # TestListTagMixin
        self._tag_list_resource_mock = self.network_client.networks

    def test_network_list_no_options(self):
        arglist = []
        verifylist = [
            ('external', False),
            ('long', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_list_external(self):
        arglist = [
            '--external',
        ]
        verifylist = [
            ('external', True),
            ('long', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{'router:external': True, 'is_router_external': True}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_list_internal(self):
        arglist = [
            '--internal',
        ]
        verifylist = [
            ('internal', True),
            ('long', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{'router:external': False, 'is_router_external': False}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
            ('external', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with()
        self.assertEqual(self.columns_long, columns)
        self.assertCountEqual(self.data_long, list(data))

    def test_list_name(self):
        test_name = "fakename"
        arglist = [
            '--name',
            test_name,
        ]
        verifylist = [
            ('external', False),
            ('long', False),
            ('name', test_name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{'name': test_name}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_enable(self):
        arglist = [
            '--enable',
        ]
        verifylist = [
            ('long', False),
            ('external', False),
            ('enable', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{'admin_state_up': True, 'is_admin_state_up': True}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_disable(self):
        arglist = [
            '--disable',
        ]
        verifylist = [('long', False), ('external', False), ('disable', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{'admin_state_up': False, 'is_admin_state_up': False}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_project(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.network_client.networks.assert_called_once_with(
            **{'project_id': project.id}
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_project_domain(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project',
            project.id,
            '--project-domain',
            project.domain_id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'project_id': project.id}

        self.network_client.networks.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_share(self):
        arglist = [
            '--share',
        ]
        verifylist = [
            ('long', False),
            ('share', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{'shared': True, 'is_shared': True}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_no_share(self):
        arglist = [
            '--no-share',
        ]
        verifylist = [
            ('long', False),
            ('no_share', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{'shared': False, 'is_shared': False}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_status(self):
        choices = ['ACTIVE', 'BUILD', 'DOWN', 'ERROR']
        test_status = random.choice(choices)
        arglist = [
            '--status',
            test_status,
        ]
        verifylist = [
            ('long', False),
            ('status', test_status),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{'status': test_status}
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_provider_network_type(self):
        network_type = self._network[0].provider_network_type
        arglist = [
            '--provider-network-type',
            network_type,
        ]
        verifylist = [
            ('provider_network_type', network_type),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{
                'provider:network_type': network_type,
                'provider_network_type': network_type,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_provider_physical_network(self):
        physical_network = self._network[0].provider_physical_network
        arglist = [
            '--provider-physical-network',
            physical_network,
        ]
        verifylist = [
            ('physical_network', physical_network),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{
                'provider:physical_network': physical_network,
                'provider_physical_network': physical_network,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_provider_segment(self):
        segmentation_id = self._network[0].provider_segmentation_id
        arglist = [
            '--provider-segment',
            segmentation_id,
        ]
        verifylist = [
            ('segmentation_id', segmentation_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{
                'provider:segmentation_id': segmentation_id,
                'provider_segmentation_id': segmentation_id,
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_network_list_dhcp_agent(self):
        arglist = ['--agent', self._agent.id]
        verifylist = [
            ('agent_id', self._agent.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.dhcp_agent_hosting_networks.assert_called_once_with(
            self._agent
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(list(data), list(self.data))

    def test_list_with_tag_options(self):
        arglist = [
            '--tags',
            'red,blue',
            '--any-tags',
            'red,green',
            '--not-tags',
            'orange,yellow',
            '--not-any-tags',
            'black,white',
        ]
        verifylist = [
            ('tags', ['red', 'blue']),
            ('any_tags', ['red', 'green']),
            ('not_tags', ['orange', 'yellow']),
            ('not_any_tags', ['black', 'white']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.networks.assert_called_once_with(
            **{
                'tags': 'red,blue',
                'any_tags': 'red,green',
                'not_tags': 'orange,yellow',
                'not_any_tags': 'black,white',
            }
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))


class TestSetNetwork(TestNetwork):
    # The network to set.
    _network = network_fakes.create_one_network({'tags': ['green', 'red']})
    qos_policy = network_fakes.create_one_qos_policy(
        attrs={'id': _network.qos_policy_id}
    )

    def setUp(self):
        super().setUp()

        self.network_client.update_network.return_value = None
        self.network_client.set_tags.return_value = None

        self.network_client.find_network.return_value = self._network

        self.network_client.find_qos_policy.return_value = self.qos_policy

        # Get the command object to test
        self.cmd = network.SetNetwork(self.app, None)

    def test_set_this(self):
        arglist = [
            self._network.name,
            '--enable',
            '--name',
            'noob',
            '--share',
            '--description',
            self._network.description,
            '--dns-domain',
            'example.org.',
            '--external',
            '--default',
            '--provider-network-type',
            'vlan',
            '--provider-physical-network',
            'physnet1',
            '--provider-segment',
            '400',
            '--enable-port-security',
            '--qos-policy',
            self.qos_policy.name,
        ]
        verifylist = [
            ('network', self._network.name),
            ('enable', True),
            ('description', self._network.description),
            ('name', 'noob'),
            ('share', True),
            ('external', True),
            ('default', True),
            ('provider_network_type', 'vlan'),
            ('physical_network', 'physnet1'),
            ('segmentation_id', '400'),
            ('enable_port_security', True),
            ('qos_policy', self.qos_policy.name),
            ('dns_domain', 'example.org.'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'name': 'noob',
            'admin_state_up': True,
            'description': self._network.description,
            'shared': True,
            'router:external': True,
            'is_default': True,
            'provider:network_type': 'vlan',
            'provider:physical_network': 'physnet1',
            'provider:segmentation_id': '400',
            'port_security_enabled': True,
            'qos_policy_id': self.qos_policy.id,
            'dns_domain': 'example.org.',
        }
        self.network_client.update_network.assert_called_once_with(
            self._network, **attrs
        )
        self.assertIsNone(result)

    def test_set_that(self):
        arglist = [
            self._network.name,
            '--disable',
            '--no-share',
            '--internal',
            '--disable-port-security',
            '--no-qos-policy',
        ]
        verifylist = [
            ('network', self._network.name),
            ('disable', True),
            ('no_share', True),
            ('internal', True),
            ('disable_port_security', True),
            ('no_qos_policy', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'admin_state_up': False,
            'shared': False,
            'router:external': False,
            'port_security_enabled': False,
            'qos_policy_id': None,
        }
        self.network_client.update_network.assert_called_once_with(
            self._network, **attrs
        )
        self.assertIsNone(result)

    def test_set_to_empty(self):
        # Test if empty strings are accepted to clear any of the fields,
        # so once they are set to a value its possible to clear them again.

        arglist = [
            self._network.name,
            '--name',
            '',
            '--description',
            '',
            '--dns-domain',
            '',
        ]
        verifylist = [
            ('network', self._network.name),
            ('description', ''),
            ('name', ''),
            ('dns_domain', ''),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'name': '',
            'description': '',
            'dns_domain': '',
        }
        self.network_client.update_network.assert_called_once_with(
            self._network, **attrs
        )
        self.assertIsNone(result)

    def test_set_nothing(self):
        arglist = [
            self._network.name,
        ]
        verifylist = [
            ('network', self._network.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network_client.update_network.called)
        self.assertFalse(self.network_client.set_tags.called)
        self.assertIsNone(result)

    def _test_set_tags(self, with_tags=True):
        if with_tags:
            arglist = ['--tag', 'red', '--tag', 'blue']
            verifylist = [('tags', ['red', 'blue'])]
            expected_args = ['red', 'blue', 'green']
        else:
            arglist = ['--no-tag']
            verifylist = [('no_tag', True)]
            expected_args = []
        arglist.append(self._network.name)
        verifylist.append(('network', self._network.name))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network_client.update_network.called)
        self.network_client.set_tags.assert_called_once_with(
            self._network, tests_utils.CompareBySet(expected_args)
        )
        self.assertIsNone(result)

    def test_set_with_tags(self):
        self._test_set_tags(with_tags=True)

    def test_set_with_no_tag(self):
        self._test_set_tags(with_tags=False)


class TestShowNetwork(TestNetwork):
    # The network to show.
    _network = network_fakes.create_one_network()
    columns = (
        'admin_state_up',
        'availability_zone_hints',
        'availability_zones',
        'created_at',
        'description',
        'dns_domain',
        'id',
        'ipv4_address_scope',
        'ipv6_address_scope',
        'is_default',
        'is_vlan_transparent',
        'is_vlan_qinq',
        'mtu',
        'name',
        'port_security_enabled',
        'project_id',
        'provider:network_type',
        'provider:physical_network',
        'provider:segmentation_id',
        'qos_policy_id',
        'router:external',
        'shared',
        'status',
        'segments',
        'subnets',
        'tags',
        'revision_number',
        'updated_at',
    )

    data = (
        network.AdminStateColumn(_network.is_admin_state_up),
        format_columns.ListColumn(_network.availability_zone_hints),
        format_columns.ListColumn(_network.availability_zones),
        _network.created_at,
        _network.description,
        _network.dns_domain,
        _network.id,
        _network.ipv4_address_scope_id,
        _network.ipv6_address_scope_id,
        _network.is_default,
        _network.mtu,
        _network.name,
        _network.is_port_security_enabled,
        _network.project_id,
        _network.provider_network_type,
        _network.provider_physical_network,
        _network.provider_segmentation_id,
        _network.qos_policy_id,
        network.RouterExternalColumn(_network.is_router_external),
        _network.is_shared,
        _network.is_vlan_transparent,
        _network.is_vlan_qinq,
        _network.status,
        _network.segments,
        format_columns.ListColumn(_network.subnet_ids),
        format_columns.ListColumn(_network.tags),
        _network.revision_number,
        _network.updated_at,
    )

    def setUp(self):
        super().setUp()

        self.network_client.find_network.return_value = self._network

        # Get the command object to test
        self.cmd = network.ShowNetwork(self.app, None)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(
            tests_utils.ParserException,
            self.check_parser,
            self.cmd,
            arglist,
            verifylist,
        )

    def test_show_all_options(self):
        arglist = [
            self._network.name,
        ]
        verifylist = [
            ('network', self._network.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network_client.find_network.assert_called_once_with(
            self._network.name, ignore_missing=False
        )

        self.assertEqual(set(self.columns), set(columns))
        self.assertCountEqual(self.data, data)


class TestUnsetNetwork(TestNetwork):
    # The network to set.
    _network = network_fakes.create_one_network({'tags': ['green', 'red']})
    qos_policy = network_fakes.create_one_qos_policy(
        attrs={'id': _network.qos_policy_id}
    )

    def setUp(self):
        super().setUp()

        self.network_client.update_network.return_value = None
        self.network_client.set_tags.return_value = None

        self.network_client.find_network.return_value = self._network

        self.network_client.find_qos_policy.return_value = self.qos_policy

        # Get the command object to test
        self.cmd = network.UnsetNetwork(self.app, None)

    def test_unset_nothing(self):
        arglist = [
            self._network.name,
        ]
        verifylist = [
            ('network', self._network.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network_client.update_network.called)
        self.assertFalse(self.network_client.set_tags.called)
        self.assertIsNone(result)

    def _test_unset_tags(self, with_tags=True):
        if with_tags:
            arglist = ['--tag', 'red', '--tag', 'blue']
            verifylist = [('tags', ['red', 'blue'])]
            expected_args = ['green']
        else:
            arglist = ['--all-tag']
            verifylist = [('all_tag', True)]
            expected_args = []
        arglist.append(self._network.name)
        verifylist.append(('network', self._network.name))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network_client.update_network.called)
        self.network_client.set_tags.assert_called_once_with(
            self._network, tests_utils.CompareBySet(expected_args)
        )
        self.assertIsNone(result)

    def test_unset_with_tags(self):
        self._test_unset_tags(with_tags=True)

    def test_unset_with_all_tag(self):
        self._test_unset_tags(with_tags=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from openstackclient.tests.functional import base


class BaseVolumeTests(base.TestCase):
    """Base class for Volume functional tests."""

    @classmethod
    def wait_for_status(
        cls,
        check_type,
        check_name,
        desired_status,
        wait=120,
        interval=5,
        failures=None,
    ):
        current_status = "notset"
        if failures is None:
            failures = ['error']
        total_sleep = 0
        while total_sleep < wait:
            output = cls.openstack(
                check_type + ' show ' + check_name,
                parse_output=True,
            )
            current_status = output['status']
            if current_status == desired_status:
                print(
                    f'{check_type} {check_name} now has status {current_status}'
                )
                return
            print(
                f'Checking {check_type} {check_name} Waiting for {desired_status} current status: {current_status}'
            )
            if current_status in failures:
                raise Exception(
                    f'Current status {current_status} of {check_type} {check_name} is one of failures {failures}'
                )
            time.sleep(interval)
            total_sleep += interval
        cls.assertOutput(desired_status, current_status)

    @classmethod
    def wait_for_delete(
        cls, check_type, check_name, wait=120, interval=5, name_field=None
    ):
        total_sleep = 0
        name_field = name_field or 'Name'
        while total_sleep < wait:
            result = cls.openstack(check_type + ' list', parse_output=True)
            names = [x[name_field] for x in result]
            if check_name not in names:
                print(f'{check_type} {check_name} is now deleted')
                return
            print(f'Checking {check_type} {check_name} Waiting for deleted')
            time.sleep(interval)
            total_sleep += interval
        raise Exception(
            f'Timeout: {check_type} {check_name} was not deleted in {wait} seconds'
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from openstackclient.common import limits
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes


class TestComputeLimits(compute_fakes.TestComputev2):
    absolute_columns = ['Name', 'Value']
    rate_columns = ["Verb", "URI", "Value", "Remain", "Unit", "Next Available"]

    def setUp(self):
        super().setUp()
        self.app.client_manager.volume_endpoint_enabled = False

        self.fake_limits = compute_fakes.create_limits()

        self.absolute_data = [
            ('floating_ips', 10),
            ('floating_ips_used', 0),
            ('image_meta', 128),
            ('instances', 10),
            ('instances_used', 0),
            ('keypairs', 100),
            ('max_image_meta', 128),
            ('max_security_group_rules', 20),
            ('max_security_groups', 10),
            ('max_server_group_members', 10),
            ('max_server_groups', 10),
            ('max_server_meta', 128),
            ('max_total_cores', 20),
            ('max_total_floating_ips', 10),
            ('max_total_instances', 10),
            ('max_total_keypairs', 100),
            ('max_total_ram_size', 51200),
            ('personality', 5),
            ('personality_size', 10240),
            ('security_group_rules', 20),
            ('security_groups', 10),
            ('security_groups_used', 0),
            ('server_group_members', 10),
            ('server_groups', 10),
            ('server_groups_used', 0),
            ('server_meta', 128),
            ('total_cores', 20),
            ('total_cores_used', 0),
            ('total_floating_ips_used', 0),
            ('total_instances_used', 0),
            ('total_ram', 51200),
            ('total_ram_used', 0),
            ('total_security_groups_used', 0),
            ('total_server_groups_used', 0),
        ]
        self.rate_data = [
            ('POST', '*', 10, 2, 'MINUTE', '2011-12-15T22:42:45Z'),
            ('PUT', '*', 10, 2, 'MINUTE', '2011-12-15T22:42:45Z'),
            ('DELETE', '*', 100, 100, 'MINUTE', '2011-12-15T22:42:45Z'),
        ]

        self.compute_client.get_limits.return_value = self.fake_limits

    def test_compute_show_absolute(self):
        arglist = ['--absolute']
        verifylist = [('is_absolute', True)]
        cmd = limits.ShowLimits(self.app, None)
        parsed_args = self.check_parser(cmd, arglist, verifylist)

        columns, data = cmd.take_action(parsed_args)

        self.assertEqual(self.absolute_columns, columns)
        self.assertEqual(self.absolute_data, data)

    def test_compute_show_rate(self):
        arglist = ['--rate']
        verifylist = [('is_rate', True)]
        cmd = limits.ShowLimits(self.app, None)
        parsed_args = self.check_parser(cmd, arglist, verifylist)

        columns, data = cmd.take_action(parsed_args)

        self.assertEqual(self.rate_columns, columns)
        self.assertEqual(self.rate_data, data)


class TestVolumeLimits(volume_fakes.TestVolume):
    absolute_columns = ['Name', 'Value']
    rate_columns = ["Verb", "URI", "Value", "Remain", "Unit", "Next Available"]

    def setUp(self):
        super().setUp()
        self.app.client_manager.compute_endpoint_enabled = False

        self.fake_limits = volume_fakes.create_limits()

        self.absolute_data = [
            ('max_total_backup_gigabytes', 1000),
            ('max_total_backups', 10),
            ('max_total_snapshots', 10),
            ('max_total_volume_gigabytes', 1000),
            ('max_total_volumes', 10),
            ('total_backup_gigabytes_used', 0),
            ('total_backups_used', 0),
            ('total_gigabytes_used', 35),
            ('total_snapshots_used', 1),
            ('total_volumes_used', 4),
        ]
        self.rate_data = [
            ('POST', '*', 10, 2, 'MINUTE', '2011-12-15T22:42:45Z'),
            ('PUT', '*', 10, 2, 'MINUTE', '2011-12-15T22:42:45Z'),
            ('DELETE', '*', 100, 100, 'MINUTE', '2011-12-15T22:42:45Z'),
        ]

        self.volume_sdk_client.get_limits.return_value = self.fake_limits

    def test_volume_show_absolute(self):
        arglist = ['--absolute']
        verifylist = [('is_absolute', True)]
        cmd = limits.ShowLimits(self.app, None)
        parsed_args = self.check_parser(cmd, arglist, verifylist)

        columns, data = cmd.take_action(parsed_args)

        self.assertEqual(self.absolute_columns, columns)
        self.assertEqual(self.absolute_data, data)

    def test_volume_show_rate(self):
        arglist = ['--rate']
        verifylist = [('is_rate', True)]
        cmd = limits.ShowLimits(self.app, None)
        parsed_args = self.check_parser(cmd, arglist, verifylist)

        columns, data = cmd.take_action(parsed_args)

        self.assertEqual(self.rate_columns, columns)
        self.assertEqual(self.rate_data, data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest.mock import call

from osc_lib import exceptions

from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import volume_message


class TestVolumeMessage(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.projects_mock = self.identity_client.projects
        self.projects_mock.reset_mock()

        self.volume_messages_mock = self.volume_client.messages
        self.volume_messages_mock.reset_mock()


class TestVolumeMessageDelete(TestVolumeMessage):
    fake_messages = volume_fakes.create_volume_messages(count=2)

    def setUp(self):
        super().setUp()

        self.volume_messages_mock.get = volume_fakes.get_volume_messages(
            self.fake_messages,
        )
        self.volume_messages_mock.delete.return_value = None

        # Get the command object to mock
        self.cmd = volume_message.DeleteMessage(self.app, None)

    def test_message_delete(self):
        self.set_volume_api_version('3.3')

        arglist = [
            self.fake_messages[0].id,
        ]
        verifylist = [
            ('message_ids', [self.fake_messages[0].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_messages_mock.delete.assert_called_with(
            self.fake_messages[0].id
        )
        self.assertIsNone(result)

    def test_message_delete_multiple_messages(self):
        self.set_volume_api_version('3.3')

        arglist = [
            self.fake_messages[0].id,
            self.fake_messages[1].id,
        ]
        verifylist = [
            ('message_ids', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for m in self.fake_messages:
            calls.append(call(m.id))
        self.volume_messages_mock.delete.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_message_delete_multiple_messages_with_exception(self):
        self.set_volume_api_version('3.3')

        arglist = [
            self.fake_messages[0].id,
            'invalid_message',
        ]
        verifylist = [
            ('message_ids', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.volume_messages_mock.delete.side_effect = [
            self.fake_messages[0],
            exceptions.CommandError,
        ]

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertEqual('Failed to delete 1 of 2 messages.', str(exc))

        self.volume_messages_mock.delete.assert_any_call(
            self.fake_messages[0].id
        )
        self.volume_messages_mock.delete.assert_any_call('invalid_message')

        self.assertEqual(2, self.volume_messages_mock.delete.call_count)

    def test_message_delete_pre_v33(self):
        self.set_volume_api_version('3.2')

        arglist = [
            self.fake_messages[0].id,
        ]
        verifylist = [
            ('message_ids', [self.fake_messages[0].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.3 or greater is required', str(exc)
        )


class TestVolumeMessageList(TestVolumeMessage):
    fake_project = identity_fakes.FakeProject.create_one_project()
    fake_messages = volume_fakes.create_volume_messages(count=3)

    columns = (
        'ID',
        'Event ID',
        'Resource Type',
        'Resource UUID',
        'Message Level',
        'User Message',
        'Request ID',
        'Created At',
        'Guaranteed Until',
    )
    data = []
    for fake_message in fake_messages:
        data.append(
            (
                fake_message.id,
                fake_message.event_id,
                fake_message.resource_type,
                fake_message.resource_uuid,
                fake_message.message_level,
                fake_message.user_message,
                fake_message.request_id,
                fake_message.created_at,
                fake_message.guaranteed_until,
            )
        )

    def setUp(self):
        super().setUp()

        self.projects_mock.get.return_value = self.fake_project
        self.volume_messages_mock.list.return_value = self.fake_messages
        # Get the command to test
        self.cmd = volume_message.ListMessages(self.app, None)

    def test_message_list(self):
        self.set_volume_api_version('3.3')

        arglist = []
        verifylist = [
            ('project', None),
            ('marker', None),
            ('limit', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        search_opts = {
            'project_id': None,
        }
        self.volume_messages_mock.list.assert_called_with(
            search_opts=search_opts,
            marker=None,
            limit=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_message_list_with_options(self):
        self.set_volume_api_version('3.3')

        arglist = [
            '--project',
            self.fake_project.name,
            '--marker',
            self.fake_messages[0].id,
            '--limit',
            '3',
        ]
        verifylist = [
            ('project', self.fake_project.name),
            ('marker', self.fake_messages[0].id),
            ('limit', 3),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        search_opts = {
            'project_id': self.fake_project.id,
        }
        self.volume_messages_mock.list.assert_called_with(
            search_opts=search_opts,
            marker=self.fake_messages[0].id,
            limit=3,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_message_list_pre_v33(self):
        self.set_volume_api_version('3.2')

        arglist = []
        verifylist = [
            ('project', None),
            ('marker', None),
            ('limit', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.3 or greater is required', str(exc)
        )


class TestVolumeMessageShow(TestVolumeMessage):
    fake_message = volume_fakes.create_one_volume_message()

    columns = (
        'created_at',
        'event_id',
        'guaranteed_until',
        'id',
        'message_level',
        'request_id',
        'resource_type',
        'resource_uuid',
        'user_message',
    )
    data = (
        fake_message.created_at,
        fake_message.event_id,
        fake_message.guaranteed_until,
        fake_message.id,
        fake_message.message_level,
        fake_message.request_id,
        fake_message.resource_type,
        fake_message.resource_uuid,
        fake_message.user_message,
    )

    def setUp(self):
        super().setUp()

        self.volume_messages_mock.get.return_value = self.fake_message
        # Get the command object to test
        self.cmd = volume_message.ShowMessage(self.app, None)

    def test_message_show(self):
        self.set_volume_api_version('3.3')

        arglist = [self.fake_message.id]
        verifylist = [('message_id', self.fake_message.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.volume_messages_mock.get.assert_called_with(self.fake_message.id)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_message_show_pre_v33(self):
        self.set_volume_api_version('3.2')

        arglist = [self.fake_message.id]
        verifylist = [('message_id', self.fake_message.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.3 or greater is required', str(exc)
        )



#   Copyright 2012 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Quota action implementations"""

import argparse
import itertools
import logging
import sys
from collections.abc import Iterable, Sequence
from typing import Any

from openstack import exceptions as sdk_exceptions
from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _

LOG = logging.getLogger(__name__)

# List the quota items, map the internal argument name to the option
# name that the user sees.

COMPUTE_QUOTAS = {
    'cores': 'cores',
    'injected_file_content_bytes': 'injected-file-size',
    'injected_file_path_bytes': 'injected-path-size',
    'injected_files': 'injected-files',
    'instances': 'instances',
    'key_pairs': 'key-pairs',
    'metadata_items': 'properties',
    'ram': 'ram',
    'server_group_members': 'server-group-members',
    'server_groups': 'server-groups',
}

VOLUME_QUOTAS = {
    'backups': 'backups',
    'backup_gigabytes': 'backup-gigabytes',
    'gigabytes': 'gigabytes',
    'per_volume_gigabytes': 'per-volume-gigabytes',
    'snapshots': 'snapshots',
    'volumes': 'volumes',
}

IMPACT_VOLUME_TYPE_QUOTAS = [
    'gigabytes',
    'snapshots',
    'volumes',
]

NETWORK_QUOTAS = {
    'floatingip': 'floating-ips',
    'security_group_rule': 'secgroup-rules',
    'security_group': 'secgroups',
    'network': 'networks',
    'subnet': 'subnets',
    'port': 'ports',
    'router': 'routers',
    'rbac_policy': 'rbac-policies',
    'subnetpool': 'subnetpools',
}

NETWORK_KEYS = [
    'floating_ips',
    'networks',
    'rbac_policies',
    'routers',
    'ports',
    'security_group_rules',
    'security_groups',
    'subnet_pools',
    'subnets',
]


def _xform_get_quota(
    data: Any,
    value: str,
    keys: Any,
) -> list[dict[str, Any]]:
    res: list[dict[str, Any]] = []
    res_info: dict[str, Any] = {}
    for key in keys:
        res_info[key] = getattr(data, key, '')

    res_info['id'] = value
    res.append(res_info)
    return res


def get_project(app: Any, project: str | None) -> dict[str, str | None]:
    if project is not None:
        identity_client = sdk_utils.ensure_service_version(
            app.client_manager.sdk_connection.identity, '3'
        )
        found_project = identity_client.find_project(
            project, ignore_missing=False
        )
        project_id = found_project.id
        project_name = found_project.name
    elif app.client_manager.auth_ref:
        # Get the project from the current auth
        auth_ref = app.client_manager.auth_ref
        project_id = auth_ref.project_id
        project_name = auth_ref.project_name
    else:
        project_id = None
        project_name = None

    return {
        'id': project_id,
        'name': project_name,
    }


def get_compute_quotas(
    app: Any,
    project_id: str | None,
    *,
    detail: bool = False,
    default: bool = False,
) -> dict[str, Any]:
    try:
        client = app.client_manager.compute
        if default:
            quota = client.get_quota_set_defaults(project_id)
        else:
            quota = client.get_quota_set(project_id, usage=detail)
    except sdk_exceptions.EndpointNotFound:
        return {}
    data: dict[str, Any] = quota.to_dict()
    if not detail:
        del data['usage']
        del data['reservation']
    return data


def get_volume_quotas(
    app: Any,
    project_id: str | None,
    *,
    detail: bool = False,
    default: bool = False,
) -> dict[str, Any]:
    try:
        client = app.client_manager.sdk_connection.volume
        if default:
            quota = client.get_quota_set_defaults(project_id)
        else:
            quota = client.get_quota_set(project_id, usage=detail)
    except sdk_exceptions.EndpointNotFound:
        return {}
    data: dict[str, Any] = quota.to_dict()
    if not detail:
        del data['usage']
        del data['reservation']
    return data


def get_network_quotas(
    app: Any,
    project_id: str | None,
    *,
    detail: bool = False,
    default: bool = False,
) -> dict[str, Any]:
    def _network_quota_to_dict(
        network_quota: Any, detail: bool = False
    ) -> dict[str, Any]:
        data: dict[str, Any] = network_quota.to_dict(computed=False)

        if not detail:
            return data

        # Neutron returns quota details in dict which is in format like:
        # {'resource_name': {'in_use': X, 'limit': Y, 'reserved': Z},
        #  'resource_name_2': {'in_use': X2, 'limit': Y2, 'reserved': Z2}}
        #
        # but Nova and Cinder returns quota in different format, like:
        # {'resource_name': X,
        #  'resource_name_2': X2,
        #  'usage': {
        #     'resource_name': Y,
        #     'resource_name_2': Y2
        #  },
        #  'reserved': {
        #     'resource_name': Z,
        #     'resource_name_2': Z2
        #  }}
        #
        # so we need to make conversion to have data in same format from
        # all of the services
        result: dict[str, Any] = {"usage": {}, "reservation": {}}
        for key, values in data.items():
            if values is None:
                continue
            if isinstance(values, dict):
                result[key] = values['limit']
                result["reservation"][key] = values['reserved']
                result["usage"][key] = values['used']

        return result

    # we have nothing to return if we are not using neutron
    if not app.client_manager.is_network_endpoint_enabled():
        return {}

    client = app.client_manager.network
    if default:
        network_quota = client.get_quota_default(project_id)
        network_quota = _network_quota_to_dict(network_quota)
    else:
        network_quota = client.get_quota(project_id, details=detail)
        network_quota = _network_quota_to_dict(network_quota, detail=detail)
    return network_quota


class ListQuota(command.Lister):
    """List quotas for all projects with non-default quota values.

    Empty output means all projects are using default quotas, which can be
    inspected with 'openstack quota show --default'.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        option = parser.add_mutually_exclusive_group(required=True)
        option.add_argument(
            '--compute',
            action='store_true',
            default=False,
            help=_('List compute quota'),
        )
        option.add_argument(
            '--volume',
            action='store_true',
            default=False,
            help=_('List volume quota'),
        )
        option.add_argument(
            '--network',
            action='store_true',
            default=False,
            help=_('List network quota'),
        )
        return parser

    def _list_quota_compute(
        self,
        parsed_args: argparse.Namespace,
        project_ids: list[str],
    ) -> tuple[tuple[str, ...], Any]:
        compute_client = self.app.client_manager.compute
        result = []

        for project_id in project_ids:
            try:
                project_data = compute_client.get_quota_set(project_id)
            # NOTE(stephenfin): Unfortunately, Nova raises a HTTP 400 (Bad
            # Request) if the project ID is invalid, even though the project
            # ID is actually the resource's identifier which would normally
            # lead us to expect a HTTP 404 (Not Found).
            except (
                sdk_exceptions.BadRequestException,
                sdk_exceptions.ForbiddenException,
                sdk_exceptions.NotFoundException,
            ) as exc:
                # Project not found, move on to next one
                LOG.warning(
                    'Project %(project_id)s not found: %(exc)s',
                    {'project_id': project_id, 'exc': exc},
                )
                continue

            project_result = _xform_get_quota(
                project_data,
                project_id,
                COMPUTE_QUOTAS.keys(),
            )

            default_data = compute_client.get_quota_set_defaults(project_id)
            default_result = _xform_get_quota(
                default_data,
                project_id,
                COMPUTE_QUOTAS.keys(),
            )

            if default_result != project_result:
                result += project_result

        columns: tuple[str, ...] = (
            'id',
            'cores',
            'injected_files',
            'injected_file_content_bytes',
            'injected_file_path_bytes',
            'instances',
            'key_pairs',
            'metadata_items',
            'ram',
            'server_groups',
            'server_group_members',
        )
        column_headers: tuple[str, ...] = (
            'Project ID',
            'Cores',
            'Injected Files',
            'Injected File Content Bytes',
            'Injected File Path Bytes',
            'Instances',
            'Key Pairs',
            'Metadata Items',
            'Ram',
            'Server Groups',
            'Server Group Members',
        )
        return (
            column_headers,
            (utils.get_dict_properties(s, columns) for s in result),
        )

    def _list_quota_volume(
        self,
        parsed_args: argparse.Namespace,
        project_ids: list[str],
    ) -> tuple[tuple[str, ...], Any]:
        volume_client = self.app.client_manager.sdk_connection.volume
        result = []

        for project_id in project_ids:
            try:
                project_data = volume_client.get_quota_set(project_id)
            except (
                sdk_exceptions.ForbiddenException,
                sdk_exceptions.NotFoundException,
            ) as exc:
                # Project not found, move on to next one
                LOG.warning(
                    'Project %(project_id)s not found: %(exc)s',
                    {'project_id': project_id, 'exc': exc},
                )
                continue

            project_result = _xform_get_quota(
                project_data,
                project_id,
                VOLUME_QUOTAS.keys(),
            )

            default_data = volume_client.get_quota_set_defaults(project_id)
            default_result = _xform_get_quota(
                default_data,
                project_id,
                VOLUME_QUOTAS.keys(),
            )

            if default_result != project_result:
                result += project_result

        columns: tuple[str, ...] = (
            'id',
            'backups',
            'backup_gigabytes',
            'gigabytes',
            'per_volume_gigabytes',
            'snapshots',
            'volumes',
        )
        column_headers: tuple[str, ...] = (
            'Project ID',
            'Backups',
            'Backup Gigabytes',
            'Gigabytes',
            'Per Volume Gigabytes',
            'Snapshots',
            'Volumes',
        )

        return (
            column_headers,
            (utils.get_dict_properties(s, columns) for s in result),
        )

    def _list_quota_network(
        self,
        parsed_args: argparse.Namespace,
        project_ids: list[str],
    ) -> tuple[tuple[str, ...], Any]:
        network_client = self.app.client_manager.network
        result = []

        for project_id in project_ids:
            try:
                project_data = network_client.get_quota(project_id)
            except (
                sdk_exceptions.NotFoundException,
                sdk_exceptions.ForbiddenException,
            ) as exc:
                # Project not found, move on to next one
                LOG.warning(
                    'Project %(project_id)s not found: %(exc)s',
                    {'project_id': project_id, 'exc': exc},
                )
                continue

            project_result = _xform_get_quota(
                project_data,
                project_id,
                NETWORK_KEYS,
            )

            default_data = network_client.get_quota_default(project_id)
            default_result = _xform_get_quota(
                default_data,
                project_id,
                NETWORK_KEYS,
            )

            if default_result != project_result:
                result += project_result

        columns: tuple[str, ...] = (
            'id',
            'floating_ips',
            'networks',
            'ports',
            'rbac_policies',
            'routers',
            'security_groups',
            'security_group_rules',
            'subnets',
            'subnet_pools',
        )
        column_headers: tuple[str, ...] = (
            'Project ID',
            'Floating IPs',
            'Networks',
            'Ports',
            'RBAC Policies',
            'Routers',
            'Security Groups',
            'Security Group Rules',
            'Subnets',
            'Subnet Pools',
        )

        return (
            column_headers,
            (utils.get_dict_properties(s, columns) for s in result),
        )

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        project_ids = [p.id for p in identity_client.projects()]
        if parsed_args.compute:
            return self._list_quota_compute(parsed_args, project_ids)
        elif parsed_args.volume:
            return self._list_quota_volume(parsed_args, project_ids)
        elif parsed_args.network:
            return self._list_quota_network(parsed_args, project_ids)

        # will never get here
        return ((), ())


class SetQuota(command.Command):
    _description = _("Set quotas for project or class")

    def _build_options_list(self) -> list[tuple[str, str, str]]:
        help_fmt = _('New value for the %s quota')
        # Compute volume and network quota options are always the same
        rets = [
            (k, v, help_fmt % v)
            for k, v in itertools.chain(
                COMPUTE_QUOTAS.items(),
                VOLUME_QUOTAS.items(),
                NETWORK_QUOTAS.items(),
            )
        ]
        return rets

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'project',
            metavar='<project/class>',
            nargs='?',
            help=_(
                'Set quotas for this project or class (name or ID) '
                '(defaults to current project)'
            ),
        )
        # TODO(stephenfin): Remove in OSC 8.0
        type_group = parser.add_mutually_exclusive_group()
        type_group.add_argument(
            '--class',
            dest='quota_class',
            action='store_true',
            default=False,
            help=_(
                '**Deprecated** Set quotas for <class>. '
                'Deprecated as quota classes were never fully implemented '
                'and only the default class is supported. '
                '(compute and volume only)'
            ),
        )
        type_group.add_argument(
            '--default',
            dest='default',
            action='store_true',
            default=False,
            help=_('Set default quotas for <project>'),
        )
        for k, v, h in self._build_options_list():
            parser.add_argument(
                f'--{v}',
                metavar=f'<{v}>',
                dest=k,
                type=int,
                help=h,
            )
        parser.add_argument(
            '--volume-type',
            metavar='<volume-type>',
            help=_('Set quotas for a specific <volume-type>'),
        )
        force_group = parser.add_mutually_exclusive_group()
        force_group.add_argument(
            '--force',
            action='store_true',
            dest='force',
            default=False,
            help=_(
                'Force quota update (only supported by compute and network)'
            ),
        )
        force_group.add_argument(
            '--no-force',
            action='store_false',
            dest='force',
            default=False,
            help=_(
                'Do not force quota update '
                '(only supported by compute and network) (default)'
            ),
        )
        # kept here for backwards compatibility/to keep the neutron folks happy
        force_group.add_argument(
            '--check-limit',
            action='store_false',
            dest='force',
            default=False,
            help=argparse.SUPPRESS,
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        if parsed_args.quota_class:
            msg = _(
                "The '--class' option has been deprecated. Quota classes were "
                "never fully implemented and the compute and volume services "
                "only support a single 'default' quota class while the "
                "network service does not support quota classes at all. "
                "Please use 'openstack quota set --default' instead."
            )
            self.log.warning(msg)

        if (
            parsed_args.quota_class or parsed_args.default
        ) and parsed_args.force:
            msg = _('--force cannot be used with --class or --default')
            raise exceptions.CommandError(msg)

        compute_kwargs = {}
        volume_kwargs = {}
        network_kwargs = {}
        compute_client = None
        volume_client = None
        network_client = None

        if self.app.client_manager.is_compute_endpoint_enabled():
            compute_client = self.app.client_manager.compute

            for k, v in COMPUTE_QUOTAS.items():
                value = getattr(parsed_args, k, None)
                if value is not None:
                    compute_kwargs[k] = value

            if compute_kwargs and parsed_args.force is True:
                compute_kwargs['force'] = parsed_args.force

        if self.app.client_manager.is_volume_endpoint_enabled():
            volume_client = self.app.client_manager.sdk_connection.volume

            for k, v in VOLUME_QUOTAS.items():
                value = getattr(parsed_args, k, None)
                if value is not None:
                    if (
                        parsed_args.volume_type
                        and k in IMPACT_VOLUME_TYPE_QUOTAS
                    ):
                        k = k + f'_{parsed_args.volume_type}'
                    volume_kwargs[k] = value

        if self.app.client_manager.is_network_endpoint_enabled():
            network_client = self.app.client_manager.network

            for k, v in NETWORK_QUOTAS.items():
                value = getattr(parsed_args, k, None)
                if value is not None:
                    network_kwargs[k] = value

        if network_kwargs:
            if parsed_args.force is True:
                # Unlike compute, network doesn't provide a simple boolean
                # option. Instead, it provides two options: 'force' and
                # 'check_limit' (a.k.a. 'not force')
                network_kwargs['force'] = True
            else:
                network_kwargs['check_limit'] = True

        if parsed_args.quota_class or parsed_args.default:
            if compute_kwargs and compute_client:
                compute_client.update_quota_class_set(
                    parsed_args.project or 'default',
                    **compute_kwargs,
                )
            if volume_kwargs and volume_client:
                volume_client.update_quota_class_set(
                    parsed_args.project or 'default',
                    **volume_kwargs,
                )
            if network_kwargs:
                sys.stderr.write(
                    "Network quotas are ignored since quota classes are not "
                    "supported."
                )

            return

        project_info = get_project(self.app, parsed_args.project)
        project = project_info['id']

        if compute_kwargs and compute_client:
            compute_client.update_quota_set(project, **compute_kwargs)
        if volume_kwargs and volume_client:
            volume_client.update_quota_set(project, **volume_kwargs)
        if network_kwargs and network_client:
            network_client.update_quota(project, **network_kwargs)


class ShowQuota(command.Lister):
    _description = _(
        """Show quotas for project or class.

Specify ``--os-compute-api-version 2.50`` or higher to see ``server-groups``
and ``server-group-members`` output for a given quota class."""
    )

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'project',
            metavar='<project>',
            nargs='?',
            help=_(
                'Show quotas for this project (name or ID) '
                '(defaults to current project)'
            ),
        )
        type_group = parser.add_mutually_exclusive_group()
        type_group.add_argument(
            '--default',
            dest='default',
            action='store_true',
            default=False,
            help=_('Show default quotas for <project>'),
        )
        type_group.add_argument(
            '--usage',
            dest='usage',
            action='store_true',
            default=False,
            help=_('Show details about quotas usage'),
        )
        service_group = parser.add_mutually_exclusive_group()
        service_group.add_argument(
            '--all',
            action='store_const',
            const='all',
            dest='service',
            default='all',
            help=_('Show quotas for all services'),
        )
        service_group.add_argument(
            '--compute',
            action='store_const',
            const='compute',
            dest='service',
            default='all',
            help=_('Show compute quota'),
        )
        service_group.add_argument(
            '--volume',
            action='store_const',
            const='volume',
            dest='service',
            default='all',
            help=_('Show volume quota'),
        )
        service_group.add_argument(
            '--network',
            action='store_const',
            const='network',
            dest='service',
            default='all',
            help=_('Show network quota'),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[tuple[Any, ...]]]:
        project_info = get_project(self.app, parsed_args.project)
        project = project_info['id']

        compute_quota_info = {}
        volume_quota_info = {}
        network_quota_info = {}

        # NOTE(stephenfin): These quota API calls do not validate the project
        # or class arguments and return what appears to be the default quota
        # values if the project or class does not exist. This is expected
        # behavior. However, we have already checked for the presence of the
        # project above so it shouldn't be an issue.
        if parsed_args.service == 'compute' or (
            parsed_args.service == 'all'
            and self.app.client_manager.is_compute_endpoint_enabled()
        ):
            compute_quota_info = get_compute_quotas(
                self.app,
                project,
                detail=parsed_args.usage,
                default=parsed_args.default,
            )

        if parsed_args.service == 'volume' or (
            parsed_args.service == 'all'
            and self.app.client_manager.is_volume_endpoint_enabled()
        ):
            volume_quota_info = get_volume_quotas(
                self.app,
                project,
                detail=parsed_args.usage,
                default=parsed_args.default,
            )

        if parsed_args.service == 'network' or (
            parsed_args.service == 'all'
            and self.app.client_manager.is_network_endpoint_enabled()
        ):
            network_quota_info = get_network_quotas(
                self.app,
                project,
                detail=parsed_args.usage,
                default=parsed_args.default,
            )

        info = {}
        if parsed_args.usage:
            info["reservation"] = compute_quota_info.pop("reservation", {})
            info["reservation"].update(
                volume_quota_info.pop("reservation", {})
            )
            info["reservation"].update(
                network_quota_info.pop("reservation", {})
            )

            info["usage"] = compute_quota_info.pop("usage", {})
            info["usage"].update(volume_quota_info.pop("usage", {}))
            info["usage"].update(network_quota_info.pop("usage", {}))

        info.update(compute_quota_info)
        info.update(volume_quota_info)
        info.update(network_quota_info)

        def _normalize_names(section: dict[str, Any]) -> None:
            # Map the internal quota names to the external ones
            # COMPUTE_QUOTAS and NETWORK_QUOTAS share floating-ips,
            # secgroup-rules and secgroups as dict value, so when
            # neutron is enabled, quotas of these three resources
            # in nova will be replaced by neutron's.
            for k, v in itertools.chain(
                COMPUTE_QUOTAS.items(),
                VOLUME_QUOTAS.items(),
                NETWORK_QUOTAS.items(),
            ):
                if not k == v and section.get(k) is not None:
                    section[v] = section.pop(k)

        _normalize_names(info)
        if parsed_args.usage:
            _normalize_names(info["reservation"])
            _normalize_names(info["usage"])

        # Remove the 'id' field since it's not very useful
        info.pop('id', None)

        # Remove the sdk-derived fields
        for field in ('location', 'name', 'force'):
            info.pop(field, None)

        if not parsed_args.usage:
            result = [{'resource': k, 'limit': v} for k, v in info.items()]
        else:
            result = [
                {
                    'resource': k,
                    'limit': v or 0,
                    'in_use': info['usage'].get(k, 0),
                    'reserved': info['reservation'].get(k, 0),
                }
                for k, v in info.items()
                if k not in ('usage', 'reservation')
            ]

        columns: tuple[str, ...] = (
            'resource',
            'limit',
        )
        column_headers: tuple[str, ...] = (
            'Resource',
            'Limit',
        )

        if parsed_args.usage:
            columns += (
                'in_use',
                'reserved',
            )
            column_headers += (
                'In Use',
                'Reserved',
            )

        return (
            column_headers,
            (utils.get_dict_properties(s, columns) for s in result),
        )


class DeleteQuota(command.Command):
    _description = _(
        "Delete configured quota for a project and revert to defaults."
    )

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_('Delete quotas for this project (name or ID)'),
        )
        option = parser.add_mutually_exclusive_group()
        option.add_argument(
            '--all',
            action='store_const',
            const='all',
            dest='service',
            default='all',
            help=_('Delete project quotas for all services (default)'),
        )
        option.add_argument(
            '--compute',
            action='store_const',
            const='compute',
            dest='service',
            default='all',
            help=_(
                'Delete compute quotas for the project '
                '(including network quotas when using nova-network)'
            ),
        )
        option.add_argument(
            '--volume',
            action='store_const',
            const='volume',
            dest='service',
            default='all',
            help=_('Delete volume quotas for the project'),
        )
        option.add_argument(
            '--network',
            action='store_const',
            const='network',
            dest='service',
            default='all',
            help=_('Delete network quotas for the project'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        identity_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        project = identity_client.find_project(
            parsed_args.project, ignore_missing=False
        )

        # compute quotas
        if parsed_args.service == 'compute' or (
            parsed_args.service == 'all'
            and self.app.client_manager.is_compute_endpoint_enabled()
        ):
            compute_client = self.app.client_manager.compute
            compute_client.revert_quota_set(project.id)

        # volume quotas
        if parsed_args.service == 'volume' or (
            parsed_args.service == 'all'
            and self.app.client_manager.is_volume_endpoint_enabled()
        ):
            volume_client = self.app.client_manager.sdk_connection.volume
            volume_client.revert_quota_set(project.id)

        # network quotas (but only if we're not using nova-network, otherwise
        # we already deleted the quotas in the compute step)
        if (
            parsed_args.service in {'all', 'network'}
            and self.app.client_manager.is_network_endpoint_enabled()
        ):
            network_client = self.app.client_manager.network
            network_client.delete_quota(project.id)

        return None
